/* wmdimp = Window Maker DIgital Music Player
 * Coypright (c) 2003 Daniele Tamino
 * This software covered by the GPL.  See COPYING file for details.
 *
 * parsecfg.h
 *
 * configuration file parser
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>

#include "parsecfg.h"
#include "debug.h"

#define ASSERT_MALLOC_RET(ptr, ret) \
    if(!(ptr)) { \
      fputs("BUY MANY MANY GB OF RAMMMMM!!!", stderr); \
      return (ret); \
    }

#define ASSERT_MALLOC(ptr) ASSERT_MALLOC_RET(ptr, 0)

/**************************************************************\
 * Option Tree                                                *
\**************************************************************/

/* N.B. only leaves have values */

inline int opt_tree_init(opt_tree* t)
{
    t->visited = 0;
    t->root = opt_tree_insert(NULL, nt_null);
    if(!t->root)
    return 0;
    return 1;
}

#ifndef NDEBUG
#define ASSERT(cond, msg) \
  if(!(cond)) { fprintf(stderr, (msg)); abort(); }
#else
#define ASSERT(cond, msg)
#endif

/* insert a note right after n */
opt_tree_node* opt_tree_insert(opt_tree_node* n, int type)
{
    opt_tree_node* nw = (opt_tree_node*) malloc(sizeof(opt_tree_node));
    if(!nw) return NULL;
    nw->type = type;
    if(type > nt_composite)
        nw->val.child = NULL;
    else if(type == nt_string)
        nw->val.s = nw->def.s = NULL;
    if(n) {
        nw->next = n->next;
        nw->parent = n->parent;
        n->next = nw;
    } else
        nw->next = nw->parent = NULL;
    return nw;
}

void opt_tree_free_node(opt_tree_node* nd)
{
    if(nd->type > nt_composite && nd->val.child != NULL) {
        opt_tree_free_node(nd->val.child);
        free(nd->val.child);
        nd->val.child = NULL;
    } else if(nd->type == nt_string) {
        if(nd->val.s && nd->val.s != nd->def.s)
            free(nd->val.s);
        if(nd->def.s)
            free(nd->def.s);
    }
    if(nd->next) {
        opt_tree_free_node(nd->next);
        free(nd->next);
        nd->next = NULL;
    }
}

inline void opt_tree_free(opt_tree* t) {
    opt_tree_free_node(t->root);
    free(t->root);
    t->root = NULL;
}

/* copy node src into node dst, and eventually copy children */
int opt_tree_dup(opt_tree_node* src, opt_tree_node* dst, opt_tree_node* dstpar)
{
    dst->type = src->type;
    dst->parent = dstpar;
    /* leave dst->next untouched; this is intentional */
    /* copy values */
    if(src->type == nt_string) {
        /* must strdup, not copy */
        if(src->def.s) {
            dst->def.s = strdup(src->def.s);
            if(!dst->def.s) return 0;
        } else
            dst->def.s = NULL;
        if(src->val.s) {
            if(src->val.s == src->def.s)
                dst->val.s = dst->def.s;
            else {
                dst->val.s = strdup(src->val.s);
                if(!dst->val.s) return 0;
            }
        } else
            dst->val.s = NULL;
    } else if(src->type > nt_composite) { /* if non-leaf, copy children */
        if(src->val.child) {
            opt_tree_node *s, *d, *prev = NULL;
            for(s = src->val.child; s != NULL; s = s->next) {
                d = (opt_tree_node*) malloc(sizeof(opt_tree_node));
                if(prev) prev->next = d;
                else     dst->val.child = d;
                if(!d) return 0;
                opt_tree_dup(s, d, dst);
                prev = d;
            }
            if(d) d->next = NULL;
        } else
            dst->val.child = NULL;
    } else {
        dst->val = src->val;
        dst->def = src->def;
    }
    return 1;
}

/**************************************************************\
 * Option Hash Table                                          *
\**************************************************************/

unsigned hash_str(const char* key)
{
    unsigned hash = 0;
    const unsigned char* p;
    for(p = (const unsigned char*) key; *p != '\0'; p++)
        hash = 31 * hash + *p;
    return hash;
}

int opt_hash_init(opt_hash* h)
{
    int i;
    for(i = 0; i < OPT_HASH_SIZE; i++)
        (*h)[i] = NULL;
    return 1;
}

/* note on ownership of allocated objects: tree is owned by the hash table,
 * so opt_hash_free will free it (and you must not). varname is duplicated
 * with strdup, so it can be freed right after use, a copy will be stored in
 * the table and freed with opt_hash_free.
 */

opt_hash_ent* opt_hash_do_find(opt_hash* ht, const char* varname, int insert)
{
    unsigned h = hash_str(varname) % OPT_HASH_SIZE;
    opt_hash_ent* ent;
    for(ent = (*ht)[h]; ent != NULL; ent = ent->next) {
        if(strcmp(varname, ent->varname) == 0)
            return ent;
    }
    if(insert) {
        ent = (opt_hash_ent*) malloc(sizeof(opt_hash_ent));
        if(!ent) return NULL;
        ent->varname = NULL;
        ent->tree = NULL;
        ent->next = (*ht)[h];
        (*ht)[h] = ent;
        return ent;
    } else
        return NULL;
}

int opt_hash_insert(opt_hash* ht, const char* varname, opt_tree* tree)
{
    opt_hash_ent* ent = opt_hash_do_find(ht, varname, 1);
    if(!ent) return 0;
    if(ent->varname)
        free(ent->varname);
    ent->varname = strdup(varname);
    if(!ent->varname) {
        free(ent);
        ASSERT_MALLOC(NULL);
    }
    if(ent->tree) {
        opt_tree_free(ent->tree);
        free(ent->tree);
    }
    ent->tree = tree;
    return 1;
}

opt_tree* opt_hash_find(opt_hash* ht, const char* varname)
{
    opt_hash_ent* ent = opt_hash_do_find(ht, varname, 0);
    if(ent)
        return ent->tree;
    else
        return NULL;
}

void opt_hash_free(opt_hash* ht)
{
    int i;
    opt_hash_ent* ent;
    for(i = 0; i < OPT_HASH_SIZE; i++) {
        for(ent = (*ht)[i]; ent != NULL; ) {
            opt_hash_ent* next = ent->next;
            free(ent->varname);
            opt_tree_free(ent->tree);
            free(ent->tree);
            free(ent);
            ent = next;
        }
        (*ht)[i] = NULL; /* so it can be reused right after */
    }
}

/* enumerating hash entries */

opt_hash_ent* opt_hash_start(opt_hash* ht, opt_hash_iterator* it)
{
    int i;
    it->ht = ht;
    for(i = 0; i < OPT_HASH_SIZE; i++)
        if((*ht)[i] != NULL)
            break;
    it->bucket = i;
    if(i < OPT_HASH_SIZE)
        it->cur = (*ht)[i];
    else
        it->cur = NULL;
    return it->cur;
}

opt_hash_ent* opt_hash_next(opt_hash_iterator* it)
{
    it->cur = it->cur->next;
    while(!it->cur) {
        it->bucket++;
        if(it->bucket < OPT_HASH_SIZE)
            it->cur = (*it->ht)[it->bucket];
        else {
            it->cur = NULL;
            break;
        }
    }
    return it->cur;
}

opt_hash_ent* opt_hash_cur(opt_hash_iterator *it)
{
    return it->cur;
}

/* lexer functions (splitting into tokens) */

struct token {
    enum token_type_values {
        tk_null = 0,
        tk_openbrace    = '{',
        tk_closebrace   = '}',
        tk_openbracket  = '[',
        tk_closebracket = ']',
        tk_comma        = ',',
        tk_equals       = '=',
        tk_semicolon    = ';',
        tk_plus         = '+',
        tk_colon        = ':',
        tk_eof          = EOF,
        tk_value
    } type;
    char * value; /* only for value types */
};

/* getc is a macro in some f**kin header, we don't want it.
   Translate it into an inline function. */
inline int do_getc(FILE* fp) { return getc(fp); }
#undef getc

/* object oriented-like structure */
struct lexer {
    const char* name;     /* file name (or source name in general) */
    int linenum;    /* current line */
    void* data;     /* generic data used by methods */
    /* methods */
    int (*getc)(struct lexer*); /* must take care of updating linenum */
    int (*peekc)(struct lexer*);
    int (*ungetc)(struct lexer*, int c);
};

void skip_whitespace (struct lexer* l)
{
    while(isspace(l->peekc(l)))
        l->getc(l);
}

/* reads a string token, possibly escaped with \ or quoted with " or ' 
 * term is a list of characters that can terminate the string
 */
static char* read_string(struct lexer* l, char* term)
{
    int c;
    int quote = 0; /* can be 0, ' or " */
    int done = 0;
    int escape = 0; /* 1 = take next character literally */
    struct vect s;

    if(!vect_init(&s, sizeof(char)))
        return 0; /* very low memory or memory corruption */
 
    while(!done) {
        c = l->getc(l);
        if(c == EOF)
            break;
        /* with " quoting, escape character is interpreted,, with ' quoting
           a \ character is taken literally */
        /* if(escaping || (quoting a non-quote char &&
           char is not \ with " quoting)) { take char literally } */
        if(escape || ( (quote && c != quote) &&
                       !(quote=='"' && c=='\\') )) {
            if(!vect_incr(&s, 1))
                return 0; /* very low memory or memory corruption */
            *_vect_get_last(&s, char) = c;
            escape = 0;
            continue;
        }
        switch(c) {
        case '\\':
            escape = 1;
            break;
        default:
            if(term) {
                char* t;
                for(t = term; *t != '\0'; t++)
                    if(c == *t) {
                        l->ungetc(l, c);
                        done = 1;
                        break;
                    }
            }
            if(!done) {
                if(c == '\'' || c == '"') {
                    if(!quote) quote = c;
                    else       quote = 0;
                } else if(isspace(c)) {
                    l->ungetc(l, c);
                    done = 1;
                } else {
                    if(!vect_incr(&s, 1))
                        return 0; /* very low memory or memory corruption */
                    *_vect_get_last(&s, char) = c;
                }
            }
        } /* end switch */
    } /* end while */
    if(!vect_incr(&s, 1))
        return 0; /* very low memory or memory corruption */
    *_vect_get_last(&s, char) = '\0';
    return (char*)(s.v);
}

static char tok_single_char_value[2] = "";
void free_token(struct token* tok)
{
    if(tok->value && tok->value != tok_single_char_value) {
        free(tok->value);
        tok->value = NULL;
    }
}

/* read a token, return 1 on success and 0 otherwise */
int read_token(struct lexer* l, struct token* tok)
{
    int c;
    free_token(tok);
    skip_whitespace(l);
    c = l->getc(l);
    switch(c) {
    case '{':
    case '}':
    case '[':
    case ']':
    case ',':
    case '=':
    case '+':
    case ':':
    case ';':
        tok->type = c;
        tok->value = tok_single_char_value;
        tok_single_char_value[0] = c;
        tok_single_char_value[1] = '\0';
        break;
    case EOF:
        tok->type = c;
        tok->value = strdup("EOF"); /* must use strdup so it can be freed */
        break;
    case '#':
        do
            c = l->getc(l);
        while(c != '\n' && c != EOF); /* skip till newline (or EOF) */
    return read_token(l, tok); /* repeat from beginning */
    default:
        tok->type = tk_value;
        l->ungetc(l, c);
        tok->value = read_string(l, "{][],=+:;");
        if(!tok->value)
            return 0;
        break;
    }
    return 1;
}

/* object oriented-like structure */
enum flags_values {
    pfl_optional = 1, /* are optional (bracketed) list allowed? */
    pfl_defaults = 2, /* are default values allowed? */
    pfl_arrays   = 4  /* are arrays allowed? */
};
struct parser {
    struct lexer* lex;
    int parse_err;  /* parse error? */
    int syntax_err; /* syntax error ? */
    int system_err; /* system errors (malloc, ...)? */
    int flags;
    void* data;
    /* methods */
    /* down/up/value: return true (1) on success and false (0) on error */
    int (*down)(struct parser*); /* handle an open brace (down in the tree) */
    int (*up)(struct parser*);   /* handle a closed brace (up in the tree) */
    int (*openbkt)(struct parser*); /* handle a '[' */
    int (*closebkt)(struct parser*); /* handle a ']' */
    int (*array)(struct parser*); /* handle a '+' */
    int (*value)(struct parser*, const char* value,
                 const char* def); /* handle a value (eventually with default) */ 
    int (*cmd)(struct parser*, const char* varname);
    int (*error)(struct parser*); /* handle a parser error */
};

/* all parsing functione take three parameters:
 * - p: parser object with all global data;
 * - tok: struct to use to store tokens;
 * - read_tok: whether it should read the next token (=1) or it is
 *   already stored in tok (=0);
 * The return value is 1 if the calling function should read the next
 * token by itself, 0 if it is already in tok.
 */

/*
 * Syntax for type definition:
 *  command ::= name '=' value_list ';'
 *
 *  value ::= atomic_value | atomic_value ":" default | struct
 *  atomic_value ::= "string" | "integer" | "real" | "boolean"
 *  default ::= <a string> | <an integer> | <a real number> | <a bool value>
 *  struct ::= "{" value_list "}"
 *  value_list ::= value | array | optional | value "," value_list
 *  value_list_noopt ::= value | array | value "," value_list
 *  optional ::= "[" value_list_noopt "]"
 *  array ::= "+" value
 *
 * 'default' is the default value for the optional atomic type described.
 * (it's intentionally admitted even for non optional values, as in case of
 * error or variable not present in the config file it may be useful).
 *
 * Syntax for config file:
 *  command ::= name '=' value_list ';'
 *  name ::= <a string>
 *  value ::= atomic_value | struct
 *  atomic_value ::= <a string> | <an integer> | <a real number> | <a bool value>
 *  struct ::= "{" value_list "}"
 *  value_list ::= value | value "," value_list
 *
 */

#define CHECK_FOR_EOF(tok) \
    if(tok->type == tk_eof) { \
    fprintf(stderr, "%s:%d: premature end of file.\n", \
                p->lex->name, p->lex->linenum); \
        return 0; \
    }

int do_value(struct parser* p, struct token *tok, int read_tok);

int do_array(struct parser* p, struct token *tok, int read_tok)
{
    int err = 0;
    if(!(p->flags & pfl_arrays)) {
        fprintf(stderr, "%s:%d: arrays (\"+ value\") are not allowed here.\n",
                p->lex->name, p->lex->linenum);
        p->parse_err++;
        err = 1;
    }
    if(read_tok)
        read_token(p->lex, tok);
    CHECK_FOR_EOF(tok);
    if(tok->type != tk_plus) {
        /* should never happen, see do_value_list */
        fprintf(stderr, "%s:%d: expected '+', found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
        return 0;
    }

    if(!err) p->array(p);

    return do_value(p, tok, 1);
}

int do_value_list(struct parser* p, struct token *tok, int read_tok, int noopt);

int do_optional(struct parser* p, struct token *tok, int read_tok, int err)
{
    int repeat = 0;
    if(!(p->flags & pfl_optional)) {
        fprintf(stderr, "%s:%d: optional (bracketed) lists are not allowed here.\n",
                p->lex->name, p->lex->linenum);
        p->parse_err++;
        err = 1;
    }
    if(read_tok)
        read_token(p->lex, tok);
    CHECK_FOR_EOF(tok);
    if(tok->type != tk_openbracket) {
        /* I think this should never happen, look in do_value_list */
        fprintf(stderr, "%s:%d: expected '[', found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
    }
    
    if(!err)
        p->openbkt(p);
    
    read_tok = 1;
    do {
        repeat = 0;
        read_tok = do_value_list(p, tok, read_tok, 1);
    
        if(read_tok)
            read_token(p->lex, tok);
        CHECK_FOR_EOF(tok);
    
        if(tok->type != tk_closebracket) {
            fprintf(stderr, "%s:%d: expected ']', found '%s' (missing ',' ?).\n",
                    p->lex->name, p->lex->linenum, tok->value);
            p->parse_err++;
            p->error(p);
            /* probably a missing comma, do as if there was one */
            repeat = 1;
            read_tok = 0;
        }
    } while(repeat);

    if(!err)
        p->closebkt(p);

    return 1;
}

int do_value_list(struct parser* p, struct token *tok, int read_tok, int noopt)
{
    do {
        if(read_tok)
            read_token(p->lex, tok);
        CHECK_FOR_EOF(tok);
    
        if(tok->type == tk_plus)
            return do_array(p, tok, 0);
        if(tok->type == tk_openbracket) {
            if(noopt) {
                fprintf(stderr, "%s:%d: cannot define an optional list into another.\n",
                        p->lex->name, p->lex->linenum);
                p->parse_err++;
                return do_optional(p, tok, 0, 1);
            } else
                return do_optional(p, tok, 0, 0);
        }

        read_tok = do_value(p, tok, 0);
        if(read_tok)
            read_token(p->lex, tok);

        if(tok->type == tk_comma)
            read_tok = 1;
    } while(tok->type == tk_comma);

    return 0;
}

int do_struct(struct parser* p, struct token *tok, int read_tok)
{
    int repeat = 0;
    if(read_tok)
        read_token(p->lex, tok);
    CHECK_FOR_EOF(tok);

    if(tok->type != tk_openbrace) {
        /* I think this should never happen, look in do_value */
        fprintf(stderr, "%s:%d: expected '{', found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
    }

    p->down(p);
    
    read_tok = 1;
    do {
        repeat = 0;
        read_tok = do_value_list(p, tok, read_tok, 0);
    
        if(read_tok)
            read_token(p->lex, tok);
        CHECK_FOR_EOF(tok);
    
        if(tok->type != tk_closebrace) {
            if(tok->type == tk_value) {
                fprintf(stderr, "%s:%d: missing ',' before '%s'.\n",
                        p->lex->name, p->lex->linenum, tok->value);
                read_tok = 0;
            } else {
                fprintf(stderr, "%s:%d: expected '}', found '%s'.\n",
                        p->lex->name, p->lex->linenum, tok->value);
                read_tok = 1;
            }
            /* probably a missing comma, do as if there was one */
            repeat = 1;
            p->error(p);
            p->parse_err++;
        }
    } while(repeat);

    p->up(p);

    return 1;
}

int do_value(struct parser* p, struct token *tok, int read_tok)
{
    if(read_tok)
        read_token(p->lex, tok);
    CHECK_FOR_EOF(tok);

    if(tok->type == tk_openbrace)
        return do_struct(p, tok, 0);
    else if(tok->type == tk_value) {
        char* value = tok->value;
        tok->value = NULL; /* don't free it! */
        read_token(p->lex, tok);
        if(tok->type != tk_colon) {
            read_tok = 0;
            p->value(p, value, NULL);
        } else {
            if(!(p->flags & pfl_defaults)) {
                fprintf(stderr, "%s:%d: default values ( : value) "
                        "are not allowed here.\n",
                        p->lex->name, p->lex->linenum);
                p->parse_err++;
                /* parse default value anyway, to check syntax */
            }
            read_token(p->lex, tok);
            if(tok->type != tk_value) {
                fprintf(stderr, "%s:%d: expected default value after ':'.\n",
                        p->lex->name, p->lex->linenum);
                read_tok = 0;
                p->value(p, value, NULL);
            } else {
                p->value(p, value, tok->value);
                read_tok = 1;
            }
        }
        free(value);
    } else{
        fprintf(stderr, "%s:%d: expected value or '{', found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
        /* ponendo questo a 0 si pu causare un ciclo infinito
           per input molto casuali */
        read_tok = 1; /* si pu discutere */
    }
    return read_tok;
}

/* parses a command in the config file */
int do_command(struct parser* p, struct token *tok, int read_tok)
{
    int ret = 1;

    if(read_tok)
        read_token(p->lex, tok);
    if(tok->type != tk_value) {
        if(tok->type == EOF) /* handle end of file gracefully */
            return 1;
        fprintf(stderr, "%s:%d: expected variable name, found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
        if(tok->type == tk_equals)
            read_tok = 0; /* missing var name, next tok is '=' */
        else
            read_tok = 1; /* invalid var name, skip completely */
        ret = 0;
        p->cmd(p, NULL);
    } else {
        p->cmd(p, tok->value);
        read_tok = 1;
    }


    if(read_tok)
        read_token(p->lex, tok);
    read_tok = 1;
    if(tok->type != tk_equals) {
        fprintf(stderr, "%s:%d: expected '=', found '%s'.\n",
                p->lex->name, p->lex->linenum, tok->value);
        p->parse_err++;
        p->error(p);
        read_tok = 0;
    }

    read_tok = do_value_list(p, tok, read_tok, 0);
    if(read_tok)
        read_token(p->lex, tok);
    read_tok = 1;
    if(tok->type != tk_semicolon) {
        fprintf(stderr, "%s:%d: missing ';' at end of value.\n",
                p->lex->name, p->lex->linenum);
        p->parse_err++;
        read_tok = 0;
    }
    return read_tok;
}

/* file lexer */

int file_getc(struct lexer* l)
{
    FILE* fp = (FILE*) l->data;
    int c = do_getc(fp);
    if(c == '\n')
        l->linenum++;
    return c;
}

int file_peekc(struct lexer* l)
{
    FILE* fp = (FILE*) l->data;
    int c = do_getc(fp);
    ungetc(c, fp);
    return c;
}

int file_ungetc(struct lexer* l, int c)
{
    FILE* fp = (FILE*) l->data;
    if(c == '\n')
        l->linenum--;
    return ungetc(c, fp);
}

int init_file_lexer(struct lexer* lex, FILE* fp, const char* filename)
{
    lex->name = filename;
    lex->linenum = 1;
    lex->data = fp;
    lex->getc = file_getc;
    lex->peekc = file_peekc;
    lex->ungetc = file_ungetc;
    return 1;
}

/* string lexer */

int str_peekc(struct lexer* l)
{
    char* s = (char*) l->data;
    int c = *s;
    if(c == '\0')
        return EOF;
    return c;
}

int str_getc(struct lexer* l)
{
    char* s = (char*) l->data;
    int c = *s;
    if(c == '\n')
        l->linenum++;
    if(c == '\0')
        return EOF;
    l->data = s + 1;    
    return c;
}

int str_ungetc(struct lexer* l, int c)
{
    char* s = (char*) l->data;
    if(c == '\n')
        l->linenum--;
    if(c != EOF) {
        s--;
/*      *s = c; */
        l->data = s;
    }   
    return c;
}

int init_string_lexer(struct lexer* lex, const char* str, char* name)
{
    lex->name = name;
    lex->linenum = 1;
    lex->data = str;
    lex->getc = str_getc;
    lex->peekc = str_peekc;
    lex->ungetc = str_ungetc;
    return 1;
}

#undef ASSERT_MALLOC
#define ASSERT_MALLOC(ptr) \
    if(!ptr) { \
    fprintf(stderr, "BUY MANY GB OF RAMMMMM!!!\n"); \
    p->system_err++; \
    return 0; \
    }

/* new type definition parser */

int null_method(struct parser* p) { (void)p; return 1; }

/* type parser */

typedef struct type_parser_data {
    opt_tree_node* node;
    opt_hash* ht;
    int skip;
} type_parser_data;

int type_insert_comp(struct parser* p, int type)
{
    type_parser_data* d = (type_parser_data*) p->data;
    opt_tree_node* nw;
    if(d->skip) return 1;
    nw = opt_tree_insert(d->node, type);
    if(!nw) { p->system_err++; return 0; }
    /* place the header */
    nw->val.child = opt_tree_insert(NULL, nt_null);
    if(!nw->val.child) { p->system_err++; return 0; }
    nw->val.child->parent = nw;
    d->node = nw->val.child;
    return 1;
}

/* handle an open brace (down in the tree) */
int type_down(struct parser* p)
{
    return type_insert_comp(p, nt_struct);
}

/* handle a closed brace or a closed bracket (up in the tree) */
int type_up(struct parser* p)
{
    type_parser_data* d = (type_parser_data*) p->data;
    opt_tree_node* n = d->node;
    if(d->skip) return 0;
    do {
        n = n->parent;
    } while(n->parent && n->type == nt_array);
    d->node = n;
    return 1;
}

/* handle an open bracket */
int type_openbkt(struct parser* p)
{
    return type_insert_comp(p, nt_optional);
}

/* handle the plus sign (array prefix) */
int type_array(struct parser* p)
{
    return type_insert_comp(p, nt_array);
}

/* handle a value */
#define WRONG_TYPE(type) \
    { \
    fprintf(stderr, "%s:%d: default value: expected " type  ", found '%s'.\n", \
        p->lex->name, p->lex->linenum, def); \
        p->syntax_err++; \
    ret = 0; \
    }

int type_value(struct parser* p, const char* value, const char* def)
{
    int ret = 0;
    type_parser_data* d = (type_parser_data*) p->data;
    char *endp;
    opt_tree_node *nw;
    if(d->skip) return 0;

    nw = opt_tree_insert(d->node, nt_error);
    if(!nw) { p->system_err++; return 0; }
    d->node = nw;

    if(strcasecmp(value, "real") == 0) {
        nw->type = nt_real;
        nw->def.r = 0.0;
        if(def) {
            float v = strtod(def, &endp);
            if(*endp != '\0') WRONG_TYPE("a real")
            else              nw->def.r = v;
        }
        nw->val.r = nw->def.r;
    } else if(strcasecmp(value, "integer") == 0) {
        nw->type = nt_integer;
        nw->def.i = 0;
        if(def) {
            long v = strtol(def, &endp, 0);
            if(*endp != '\0') WRONG_TYPE("an integer")
            else              nw->def.i = v;
        }
        nw->val.i = nw->def.i;
    } else if(strcasecmp(value, "boolean") == 0) {
        nw->type = nt_boolean;
        nw->def.b = 0;
        if(def) {
            int v = strtol(def, &endp, 10);
            if(*endp != '\0') {
                if(!strcasecmp(def, "yes") || !strcasecmp(def, "true"))
                    nw->def.b = 1;
                else if(!strcasecmp(def, "no") || !strcasecmp(def, "false"))
                    nw->def.b = 0;
                else
                    WRONG_TYPE("a boolean")
                } else
                    nw->def.b = v ? 1 : 0;
        }
        nw->val.b = nw->def.b;
    } else if(strcasecmp(value, "string") == 0) {
        nw->type = nt_string;
        if(def) {
            nw->def.s = strdup(def);
            if(!nw->def.s) {
                nw->def.s = NULL;
                ASSERT_MALLOC(NULL);
            }
        } else
            nw->def.s = NULL;
        nw->val.s = nw->def.s;
    } else {
        fprintf(stderr, "%s: invalid atomic type '%s.\n", p->lex->name, value);
        p->syntax_err++;
        ret = 0;
    }
    return ret;
}

int type_cmd(struct parser* p, const char* varname)
{
    type_parser_data *d = (type_parser_data*) p->data;
    opt_tree* tree;
    if(varname == NULL) {
        d->skip=1;
        return 1;
    } else
        d->skip=0;
    tree = (opt_tree*) malloc(sizeof(opt_tree));
    if(!tree) { p->system_err++; return 0; }
    if(!opt_tree_init(tree)) { p->system_err++; free(tree); return 0; }
    if(!opt_hash_insert(d->ht, varname, tree)) {
        p->system_err++; opt_tree_free(tree); free(tree);
        return 0;
    }
    d->node = tree->root;
    return 1;
}

int init_type_parser(struct parser* p, struct lexer* lex, opt_hash* ht)
{
    type_parser_data *data;
    data = (type_parser_data*)malloc(sizeof(type_parser_data));
    if(!data) {
        p->system_err++;
        return 0;
    }
    p->lex = lex;
    p->parse_err = 0;  /* parse errors ? */
    p->syntax_err = 0; /* syntax errors ? */
    p->system_err = 0; /* system errors ? */
    p->flags = (pfl_optional | pfl_arrays | pfl_defaults);
    data->ht = ht;
    data->node = NULL;
    data->skip = 0;
    p->data = data;
    /* methods */
    p->down     = type_down;
    p->up       = type_up;
    p->openbkt  = type_openbkt;
    p->closebkt = type_up;
    p->array    = type_array;
    p->value    = type_value;
    p->cmd      = type_cmd;
    p->error    = null_method;
    return 1;
}

void free_type_parser(struct parser* p)
{
    if(p->data)
    free(p->data);
}


/* new value parser */

enum val_parser_flags {
    VPFL_IGNUNKNOWN = CFL_IGNUNKNOWN, /* so they are compatible :) */
    VPFL_IGNOVERWR  = CFL_IGNOVERWR
};

typedef struct val_parser_data {
    opt_hash* ht; /* hash table containing precompiled trees */
    char* varname; /* name of current variable (even in case of error) */
    int flags;     /* load time flags (only 1 by now)  */
    int must_free; /* true if we must free varname */
    opt_tree_node* tree; /* current tree node, NULL on error */
    int skip_struct; /* struct level to skip due to errors */
} val_parser_data;


const char* get_type_name(int type)
{
    switch(type) {
    case nt_real:    return "real";
    case nt_integer: return "integer";
    case nt_boolean: return "boolean";
    case nt_string:  return "string";
    }
    return "(unknown)";
}

void clean_array(opt_tree_node* n)
{
    opt_tree_node* templ;
    opt_tree_node* last=NULL;
#ifndef NDEBUG
    if(n->type != nt_array) {
        fprintf(stderr, "BUG: incr_array called for a non-array node!");
        return;
    }
#endif
    /* search the the template (last element) and the last non template */
    for(templ = n->val.child; templ->next; templ = templ->next)
        last = templ;
    /* unlink the template */
    last->next = NULL;
    /* free everything except the template and the header */
    opt_tree_free_node(n->val.child->next);
    /* restore the template */
    n->val.child->next = templ;
}

/* returns the array template child, wiping out every other
 * child, except if this is a top level array; in this case
 * append to the end.
 * reminder: last child is the template, first is the header
 */
opt_tree_node* setup_array(opt_tree_node* n)
{
    opt_tree_node* ch;
    ASSERT(n->type==nt_array, "BUG: setup_array called with non array node!");
    ch = n->val.child;
    if(ch->next && ch->next->next) {
        if(n->parent == NULL) { /* top level array, add elements */
            while(ch->next->next) /* move to the before-last element */
                ch = ch->next;
        } else {
            clean_array(n);
            ch = n->val.child;
        }
    }
    return ch;
}

int val_down(struct parser* p) /* handle '{' */
{
    int ret=1;
    val_parser_data* d = (val_parser_data*) p->data;
    opt_tree_node* n;
    if(d->skip_struct) {
        d->skip_struct++;
        return 1;
    }
    n = d->tree = d->tree->next;

    if(n->parent && n->parent->type == nt_array && n->next == NULL) {
        opt_tree_node* nw = opt_tree_insert(n, nt_error);
        ASSERT_MALLOC(nw);
        ASSERT_MALLOC(opt_tree_dup(n, nw, n->parent));
    }

    if(n->type > nt_composite) {
    if(n->type == nt_optional) {
        /*d->optional++;*/ /* >>>TODO: is it necessary? */
        d->tree = n->val.child;
        return val_down(p);
    } else if(n->type == nt_array) {
        d->tree = setup_array(n);
        return val_down(p);
    } else
        d->tree = n->val.child;
    } else {
        fprintf(stderr, "%s:%d: expected '%s', found a structure (skipping).\n",
                p->lex->name, p->lex->linenum, get_type_name(n->type));
        d->skip_struct++;
        p->syntax_err++;
        ret = 0;
    }

    return ret;
}

int val_up(struct parser* p) /* handle '}' */
{
    int ret=1;
    val_parser_data* d = (val_parser_data*) p->data;
    opt_tree_node* n;

    if(d->skip_struct) {
        d->skip_struct--;
        return 1;
    }
    n = d->tree;

    while(n->parent->type == nt_optional || n->parent->type == nt_array)
        n = n->parent;

    /* an if should be enough, but it's just to be sure it works */
    while(n->next && n->next->type == nt_optional)
        n = n->next;
    if(n->next != NULL) {
        fprintf(stderr, "%s:%d: missing non-optional values in struct.\n",
                p->lex->name, p->lex->linenum);
        p->syntax_err++;
        ret = 0;
    }
    d->tree = n->parent;
    return ret;
}

#undef WRONG_TYPE
#define WRONG_TYPE(type) \
    { \
    fprintf(stderr, "%s:%d: expected " type  ", found '%s'.\n", \
        p->lex->name, p->lex->linenum, value); \
        p->syntax_err++; \
    ret = 0; \
    }

int val_value(struct parser* p, const char* value, const char* def)
{
    int ret=1;
    char* endp;
    val_parser_data* d = (val_parser_data*) p->data;
    opt_tree_node* n;

    if(d->skip_struct)
        return 1;

    if(d->tree->next == NULL) {
        fprintf(stderr, "%s:%d: extra value '%s' at end of struct (ignored).\n",
                p->lex->name, p->lex->linenum, value);
        p->syntax_err++;
        return 0;
    }

    n = d->tree = d->tree->next;

    if(n->parent && n->parent->type == nt_array && n->next == NULL) {
        opt_tree_node* nw = opt_tree_insert(n, nt_error);
        ASSERT_MALLOC(nw);
        ASSERT_MALLOC(opt_tree_dup(n, nw, n->parent));
    }

    switch(n->type) {
    case nt_real: {
        cfg_real val = strtod(value, &endp);
        if(*endp != '\0') WRONG_TYPE("a real")
        else n->val.r = val;
        break;
    }
    case nt_integer: {
        cfg_integer val = strtol(value, &endp, 0);
        if(*endp != '\0') WRONG_TYPE("an integer")
        else n->val.i = val;
        break;
    }
    case nt_boolean: {
        cfg_real val = strtol(value, &endp, 0);
        if(*endp != '\0') {
            if(!strcasecmp(value, "yes") || !strcasecmp(value, "true"))
                n->val.b = 1;
            else if(!strcasecmp(value, "no") || !strcasecmp(value, "false"))
                n->val.b = 0;
            else
                WRONG_TYPE("a boolean")
            }
        else n->val.b = (val != 0);
        break;
    }
    case nt_string: {
        n->val.s = strdup(value);
        ASSERT_MALLOC(n->val.s);
        break;
    }
    case nt_optional:
        d->tree = n->val.child;
        return val_value(p, value, def);
    case nt_array:
        d->tree = setup_array(n);
        return val_value(p, value, def);
    case nt_struct:
        fprintf(stderr, "%s:%d: expected a struct, found '%s'.\n",
                p->lex->name, p->lex->linenum, value);
        break;
    default:
        break;
    }
    return ret;
}

int val_cmd(struct parser* p, const char* varname)
{
    int ret = 1;
    val_parser_data* d = (val_parser_data*) p->data;
    d->skip_struct = 0;
    if(d->must_free) {
        free(d->varname);
        d->varname = NULL;
    }
    if(varname) {
        opt_tree* t = opt_hash_find(d->ht, varname);
        if(!t) {
            if(!(d->flags & VPFL_IGNUNKNOWN)) {
                fprintf(stderr, "%s:%d: unknown variable name '%s'.\n",
                        p->lex->name, p->lex->linenum, varname);
                p->syntax_err++;
            }
            d->varname = NULL;
            d->tree = NULL;
        } else {
            d->tree = t->root;
            /* check if var was already assigned a value */
            if(!(d->flags & VPFL_IGNOVERWR) && t->visited &&
               d->tree->next && d->tree->next->type != nt_array) {
                fprintf(stderr, "%s:%d: warning: new value of '%s' "
                        "will override previous.\n",
                        p->lex->name, p->lex->linenum, varname);
            }
            t->visited = 1;
            d->varname = strdup(varname);
            if(!d->varname) p->system_err++;
            else d->must_free = 1;
        }
    } else
        d->varname = NULL;
    if(!d->varname) { /* handle malloc errors and syntax errors */
        d->varname = "(no name)";
        d->must_free = 0;
        d->skip_struct++; /* skip everything */
        ret = 0;
    }
    return 1;
}

int init_val_parser(struct parser* p, struct lexer* lex, opt_hash* ht,
                    int flags)
{
    val_parser_data* data;
    data = (val_parser_data*) malloc(sizeof(val_parser_data));
    ASSERT_MALLOC(data);
    /* inizialize data to all null's */
    data->ht = ht;
    data->varname = NULL;
    data->flags = flags;
    data->must_free = 0;
    data->tree = NULL;
    data->skip_struct = 0;
    /* inizialize parser */
    p->lex = lex;
    p->parse_err = 0;  /* parse errors ? */
    p->syntax_err = 0; /* syntax errors ? */
    p->system_err = 0; /* system errors ? */
    p->flags = 0; /* no optional lists, arrays or default values */
    p->data = data;
    /* store parser methods */
    p->down     = val_down;
    p->up       = val_up;
    p->openbkt  = null_method;
    p->closebkt = null_method;
    p->array    = null_method;
    p->value    = val_value;
    p->cmd      = val_cmd;
    p->error    = null_method;
    return 1;
}

void free_val_parser(struct parser* p)
{
    if(p->data) {
        val_parser_data* d = (val_parser_data*) p->data;
        if(d->must_free)
            free(d->varname);
        free(p->data);
    }
}

int parseit(struct parser *p)
{
    struct token tok = { tk_null, NULL };
    int read_tok = 1;
    int errors;

    while(tok.type != EOF) {
        read_tok = do_command(p, &tok, read_tok);
    }
    free_token(&tok);
    errors = p->parse_err + p->syntax_err + p->system_err;
    if(errors) {
        fprintf(stderr, "%s: ", p->lex->name);
        if(p->system_err)
            fprintf(stderr, "%d system errors: you are low on memory or what?\n\t",
                    p->system_err);
        if(p->parse_err)
            fprintf(stderr, "%d parse errors, ", p->parse_err);
        if(p->syntax_err)
            fprintf(stderr, "%d syntax errors, ", p->syntax_err);
        fprintf(stderr, "%d total errors.\n", errors);
    }

    if(p->system_err)
        return CFGERR_SYSTEM;
    if(p->parse_err || p->syntax_err)
        return CFGERR_FILE;
    return CFGERR_NONE;
}

int precompile_specs(const type_desc_spec* spec, opt_hash* ht)
{
    int ret;
    struct lexer lex;
    struct parser p;
    if(!init_string_lexer(&lex, spec->spec, spec->name))
        return 0;
    if(!init_type_parser(&p, &lex, ht))
        return 0;
    ret = parseit(&p);
    free_type_parser(&p);
    if(ret == CFGERR_FILE)
        ret = CFGERR_SPECS;
    return ret;
}

int parse_config_file(FILE* fp, const char* filename, opt_hash* ht, int flags)
{
    int ret;
    struct lexer lex;
    struct parser p;
    if(!init_file_lexer(&lex, fp, filename))
        return 0;
    if(!init_val_parser(&p, &lex, ht, flags))
        return 0;
    ret = parseit(&p);
    free_val_parser(&p);
    return ret;
}

/* main loading function */
int load_config_file_multi(const char* name, const type_desc_spec** specs,
                           opt_hash* ht, int flags)
{
    struct passwd *pw;
    char *config_filename = NULL;
    FILE *fp = NULL;
    int ret = CFGERR_NONE;
    int namelen = strlen(name);

    while(*specs != NULL) {
        ret = precompile_specs(*specs, ht);
        if(ret != CFGERR_NONE) {
            fprintf(stderr, "BUG: internal parse errors. Contact author.\n");
            return ret;
        }
        specs++;
    }

    /* >>> TODO: don't know it it is good or bad to check for the file as is
     * >>> instead of searching *only* in the standard directories.
     */
    if(flags & CFL_SEARCHWD) {
        config_filename = name;
        fp = fopen(config_filename, "r");
    }

    if(!fp /* && errno == ENOENT */) {

        pw = getpwuid(getuid());
        /* don't forget about the string terminator... */
        config_filename = (char *) malloc(sizeof(char) *
                                          (strlen(pw->pw_dir) + namelen + 3));
        if(!config_filename) {
            errno = ENOMEM;
            return CFGERR_SYSTEM;
        }
        sprintf(config_filename, "%s/.%s", pw->pw_dir, name);

        fp = fopen(config_filename, "r");

        if(!fp && errno == ENOENT) {
            /* try with "/usr/local/etc/<name>" */
            free (config_filename);
            config_filename = (char *) malloc(sizeof(char) * (namelen + 16));
            if(!config_filename) {
                errno = ENOMEM;
                return CFGERR_SYSTEM;
            }
            sprintf(config_filename, "/usr/local/etc/%s", name);
            fp = fopen(config_filename, "r");
            if(!fp && errno == ENOENT) {
                /* try with "/etc/<name>" */
                sprintf(config_filename, "/etc/%s", name);
                fp = fopen(config_filename, "r");
            }
        }
    }
    
    if(!fp)
        return (errno == ENOENT) ? CFGERR_NOTFOUND : CFGERR_SYSTEM;

    /*     fprintf(stderr, "DBG: config_filename = '%s'\n", config_filename); */

    ret = parse_config_file(fp, config_filename, ht, flags);

    fclose(fp);
    if(name != config_filename)
        free(config_filename);

    return ret;
}

int load_config_file(const char* name, const char* specs, opt_hash* ht,
                     int flags)
{
    const type_desc_spec sp = { "(internal specs)", specs };
    const type_desc_spec* s[2] = { &sp, NULL };
    return load_config_file_multi(name, s, ht, flags);
}

const char* get_cfg_err(int err)
{
    switch(err) {
    case CFGERR_NONE:
        return "no error";
    case CFGERR_NOTFOUND:
        return "config file not found in ~/, /usr/local/etc or /etc";
    case CFGERR_SYSTEM:
        return /*strerror(errno)*/"there were system errors";
    case CFGERR_SPECS:
        return "error(s) parsing internal specs";
    case CFGERR_FILE:
        return "error(s) parsing configuration file";
    }
    return "(unknown/invalid error code)";
}
