/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>.
 * This software covered by the GPL.  See COPYING file for details.
 *
 * player_mad.c
 *
 * mp3 file player using libmad and output.c to play sound.
 *
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE /* to get prototype for vasprintf */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#include <unistd.h>
#include <pthread.h>
#include <fcntl.h>
#ifdef HAVE_MMAP
#  include <sys/mman.h>
#endif

#include <mad.h>
#include <id3tag.h>
#include "player_mad.h"
#include "output.h"
#include "str.h"
#include "utils.h"
#include "debug.h"

int mad_static_init(opt_hash* ht);
void mad_static_finish(void);
int mad_init(player_mad* me);
void mad_finish(struct player*);
player* mad_create(void);
int mad_test(struct player*, const char* fn);
int mad_setup(struct player*, const char* fn, char ** title, int play);
int mad_play(struct player*);
int mad_play_sync(struct player*);
int mad_pause(struct player*);
int mad_stop(struct player*);
int mad_jump_frames(struct player*, long f);
int mad_jump_time(struct player*, long sec);
int mad_jump(struct player*, long position, long total);
int mad_get_length(struct player*, song_pos* len);
int mad_get_pos(struct player*, song_pos* pos);
void mad_set_callback(struct player*, player_cb_func cb, void* param);

static player player_mad_tmpl = {
    "MAD mp3 player",
    st_stop,
    mad_finish,
    mad_test,
    mad_setup,
    mad_play,
    mad_play_sync,
    mad_pause,
    mad_stop,
    mad_jump_frames,
    mad_jump_time,
    mad_jump,
    mad_get_length,
    mad_get_pos,
    mad_set_callback
};

player_factory player_mad_factory = {
    mad_static_init,
    mad_static_finish,
    mad_create,
    {
        "mad player",
        "mad_title = string:'%t - %a';"
        "mad_title_notag = string:'%f';"
        "mad_nocrc = boolean:false;"
        "mad_halfrate = boolean:false;"
        "mad_mmap = boolean:true;"
    }
};
static char* cfg_title = NULL;
static char* cfg_title_notag = NULL;
static int cfg_nocrc = 0;
static int cfg_halfrate = 0;
static int cfg_mmap = 1;

int mad_static_init(opt_hash* ht)
{
    opt_tree* t;    

    t = opt_hash_find(ht, "mad_title");
    if(t) {
        assert(t->root && t->root->next);
        cfg_title = strdup(t->root->next->val.s);
    }
    t = opt_hash_find(ht, "mad_title_notag");
    if(t) {
        assert(t->root && t->root->next);
        cfg_title_notag = strdup(t->root->next->val.s);
    }
    t = opt_hash_find(ht, "mad_nocrc");
    if(t) {
        assert(t->root && t->root->next);
        cfg_nocrc = t->root->next->val.b;
    }
    t = opt_hash_find(ht, "mad_halfrate");
    if(t) {
        assert(t->root && t->root->next);
        cfg_halfrate = t->root->next->val.b;
    }
    t = opt_hash_find(ht, "mad_mmap");
    if(t) {
        assert(t->root && t->root->next);
        cfg_mmap = t->root->next->val.b;
    }
    DBGMSG("halfrate = %d, nocrc = %d, %s\n", cfg_halfrate, cfg_nocrc,
           cfg_mmap ? "mmap" : "no mmap");
    output_init(ht); /* >>>TODO: this will go away when multiple
                        drivers will use the output module */
    return 1;
}

/* releases static structures */
void mad_static_finish(void)
{
    if(cfg_title) free(cfg_title);
    if(cfg_title_notag) free(cfg_title_notag);
    output_finish(); /* >>>TODO: this will go away when multiple
                        drivers will use the output module */
}

/* malloc's and init's a new player of the correct type */
player* mad_create(void)
{
    player_mad* pl = (player_mad*) malloc(sizeof(player_mad));
    if(!pl) return NULL;
    mad_init(pl);

    return &pl->pl;
}

typedef struct decmsg
{
    int msg;
    int* ack;
} decmsg;

typedef struct decmsg_seek
{
    decmsg m;
    song_pos seek;
} decmsg_seek;

typedef enum wait_flag { wait_ack, dont_wait } wait_flag;
static int send_dec_msg(player_mad* me, int msg, enum wait_flag w)
{
    int ack = 0;
    decmsg* m = malloc(sizeof(decmsg));
    assert(m != NULL);
    m->msg = msg;
    m->ack = (w == wait_ack) ? &ack : NULL;
    if(thrcomm_send(&me->dec_thr_chn, m)) {
        if(w == wait_ack) {
            thrcomm_waitack(&me->dec_thr_chn, &ack);
            return ack;
        } else
            return 0;
    } else {
        free(m);
        return -1;
    }
}

static int send_dec_seek(player_mad* me, song_pos* seek, wait_flag w)
{
    int ack = 0;
    decmsg_seek* m = malloc(sizeof(decmsg_seek));
    assert(m != NULL);
    m->m.msg = MSG_SEEK;
    m->m.ack = (w == wait_ack) ? &ack : NULL;
    m->seek = *seek;
    thrcomm_send(&me->dec_thr_chn, m);
    if(w == wait_ack) {
        thrcomm_waitack(&me->dec_thr_chn, &ack);
        return ack;
    } else
        return 0;
}

static int auto_send_msg(player_mad* me, int msg)
{
    decmsg *m = malloc(sizeof(decmsg));
    if(m) {
        m->msg = msg;
        m->ack = NULL;
        /* must use trysend since this is the receiver thread */
        return thrcomm_trysend(&me->dec_thr_chn, m);
    }
    return 0;
}

static int auto_send_seek(player_mad* me, song_pos* seek)
{
    decmsg_seek *m = malloc(sizeof(decmsg_seek));
    if(m) {
        m->m.msg = MSG_SEEK;
        m->m.ack = NULL;
        m->seek = *seek;
        return thrcomm_trysend(&me->dec_thr_chn, m);
    }
    return 0;
}

typedef struct decoder {
    /* mad structures (use low-level interface) */
    struct mad_stream stream;
    struct mad_frame frame;
    struct mad_synth synth;    
} decoder;

static inline void decoder_init(decoder* d)
{
    mad_stream_init(&d->stream);
    mad_frame_init(&d->frame);
    mad_synth_init(&d->synth);
    mad_stream_options(&d->stream, (cfg_nocrc ? MAD_OPTION_IGNORECRC : 0)
                       | (cfg_halfrate ? MAD_OPTION_HALFSAMPLERATE : 0));
}

static inline void decoder_finish(decoder* d)
{
    mad_synth_finish(&d->synth);
    mad_frame_finish(&d->frame);
    mad_stream_finish(&d->stream);
}

/* scan_thread_mtx is used to let scan_thread run only when decoder_thread
   is doing nothing (i.e. waiting for audio to be written), so it shouldn't
   cause buffer underruns from one song to the next. */
static pthread_mutex_t scan_thread_mtx = PTHREAD_MUTEX_INITIALIZER;
/* signalled when scan_thread terminates (cleanly) */
static pthread_cond_t term_scan_cond = PTHREAD_COND_INITIALIZER;

typedef struct buffer {
    char* data; /* address of allocated space */
    long size; /* allocated size */
    char* ptr; /* start of unused (not yet decoded) data  */
    long len;  /* n. of bytes actually used */
} buffer;

/* load new data into buffer, return 1 on success, 0 on error (see errno)
   and -1 on end of file */
int refill_buffer(struct mad_stream* stream, buffer* buf, int fd, int* eof)
{
    ssize_t n;
    assert(buf->ptr >= buf->data);
    assert(buf->ptr + buf->len <= buf->data + buf->size);
    
    if(buf->len == buf->size)
        return 1; /* buffer already full */
    if(buf->data != buf->ptr) {
        memmove(buf->data, buf->ptr, buf->len);
        buf->ptr = buf->data;
    }
    if(*eof) {
        n = (buf->size - buf->len < MAD_BUFFER_GUARD) ?
            buf->size - buf->len : MAD_BUFFER_GUARD;
        memset(buf->ptr + buf->len, 0, n);
    } else {
        do
            n = read(fd, buf->ptr + buf->len, buf->size - buf->len);
        while(n < 0 && errno == EINTR);
        if(n < 0) return 0;
        else if(n == 0) {
            *eof = 1;
            return -1; /* EOF */
        }
    }
    buf->len += n;
    mad_stream_buffer(stream, buf->ptr, buf->len);
    return 1;
}

static inline void
send_error_vmsg(player_mad* me, const char* fmt, va_list vl)
{
    if(me->callback) {
        char* msg = NULL;
        if(vasprintf(&msg, fmt, vl) < 0) {
            me->callback(&me->pl, CB_ERROR, me->cb_param, "VERY LOW MEMORY");
        } else {
            me->callback(&me->pl, CB_ERROR, me->cb_param, msg);
            free(msg);
        }
    }
}

static inline void
send_error_msg(player_mad* me, const char* fmt, ...)
{
    va_list vl;
    va_start(vl, fmt);
    send_error_vmsg(me, fmt, vl);
    va_end(vl);
}

#if 0

void stop_on_error(player_mad* me, const char* fmt, ...)
{
    va_list vl;
    va_start(vl, fmt);
    pthread_mutex_unlock(&scan_thread_mtx);
    send_error_vmsg(me, fmt, vl);
    va_end(vl);
    output_stop(1, 0);
    me->pl.status = st_stop;
    if(me->callback)
        me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
    pthread_mutex_lock(&scan_thread_mtx);
}

#endif

#ifndef DEBUG
static const long buffer_size = 40000 * 5; /* 5 s at 320 kbps */
#else
/* use a lower value to test some error conditions */
static const long buffer_size = 20000;
#endif

/* pre-scanning thread (to get song_len and potentially every other info) */

char* get_title(player_mad* me, const char* fn);

void* scan_thread(void* pme)
{
    player_mad* me = (player_mad*) pme;
    int fd = -1;
    struct mad_stream stream;
    struct mad_header header;
    struct buffer inbuf = { NULL, 0, NULL, 0 };
    int done = 0, eof = 0, mmap = 0;
    mad_timer_t elapsed = mad_timer_zero;
    long nframes = 0;
#ifdef HAVE_INT64_T
    typedef int64_t bitrate_t;
#else
    typedef double bitrate_t;
#endif
    bitrate_t bitrate = 0;
    char* fn = strdup(me->filename);
/*  me->scan_thr_running++; */
    DBGMSG("scan_thread started for %s (pid %d)\n", fn, getpid());
    DBG(if(me->scan_thr_running > 1) {
        DBGMSG("scan_thread: multiple scan threads! starting debugger...\n");
        dbg_start_debugger();
    })
    /* >>>TODO: add support for xing headers */
    
    mad_stream_init(&stream);
    mad_header_init(&header);

#ifdef HAVE_MMAP    
    pthread_mutex_lock(&me->infile_mtx);
    if((mmap = me->mmap) != 0 && !me->abort_scan)
        mad_stream_buffer(&stream, me->buffer, me->flen);
    pthread_mutex_unlock(&me->infile_mtx);
    if(!mmap) {
#endif
        inbuf.data = malloc(buffer_size);
        if(!inbuf.data) {
            if(me->callback)
                me->callback(&me->pl, CB_ERROR, me->cb_param,
                             "scan_thread: out of memory!");
            goto end;
        }
        inbuf.ptr = inbuf.data;
        inbuf.size = buffer_size;

        pthread_mutex_lock(&me->infile_mtx);
        fd = open(fn, O_RDONLY);
        pthread_mutex_unlock(&me->infile_mtx);
        if(fd < 0) {
            send_error_msg(me, "scan_thread (open): %s", strerror(errno));
            goto end;
        }
#ifdef HAVE_MMAP
    }
#endif

    while(!done && !me->abort_scan) {
        pthread_mutex_lock(&scan_thread_mtx); /* this is to spread cpu usage */
        pthread_mutex_unlock(&scan_thread_mtx);

#ifdef HAVE_MMAP
        if(!mmap) {
#endif
            if(stream.next_frame) {
                inbuf.ptr = stream.next_frame;
                inbuf.len = inbuf.data + inbuf.size - inbuf.ptr;
            }
            if(eof || refill_buffer(&stream, &inbuf, fd, &eof) == 0)
                break;
#ifdef HAVE_MMAP
        }
#endif

        if(mmap) pthread_mutex_lock(&me->infile_mtx);
        while(!me->abort_scan) {
            int ret, id3 = 0;
            ret = mad_header_decode(&header, &stream);
            if(ret < 0) {
                if(stream.error == MAD_ERROR_BUFLEN)
                    break;
                if(stream.error == MAD_ERROR_LOSTSYNC) {
                    /* ignore LOSTSYNC due to ID3 tags */
                    int tagsize = id3_tag_query (stream.this_frame, 
                                         stream.bufend-stream.this_frame);
                    if (tagsize > 0) {
                        DBGMSG("scan_thread(%s): frame %d: ignoring %d bytes of "
                               "ID3 tag\n", fn, nframes+1, tagsize);
                        mad_stream_skip(&stream, tagsize);
                        /*pthread_mutex_unlock(&scan_thread_mtx);*/
                        id3 = 1;
                    }
                }
                if(!id3)
                    send_error_msg(me, "scan(%s): frame %d: %s", fn, nframes+1,
                                   mad_stream_errorstr(&stream));
            } else {
                mad_timer_add(&elapsed, header.duration);
                bitrate += (bitrate_t) header.bitrate;
                nframes++;
            }
        }
#ifdef HAVE_MMAP
        if(mmap) {
            pthread_mutex_unlock(&me->infile_mtx);
            break; /* EOF */
        }
#endif
    }
    if(!me->abort_scan) {
        pthread_mutex_unlock(&scan_thread_mtx);
        me->len.frames = nframes;
        me->len.seconds = elapsed.seconds +
            (float)elapsed.fraction / MAD_TIMER_RESOLUTION;
        if(nframes)
            me->bitrate = (int)(bitrate / nframes);
    }
#ifndef NDEBUG
    if(me->abort_scan)
        DBGMSG("scan_thread(%s): aborted\n", fn);
    else
        DBGMSG("scan_thread(%s): %ld frames, %g seconds, %d Kbit/sec\n",
               fn, nframes, me->len.seconds, me->bitrate);
#endif
 end:
    if(!me->abort_scan && strstr(cfg_title, "%b") != NULL) {
        char* s;
        /* bitrate is known only in this moment, so reformat the title */
        pthread_mutex_lock(&me->title_mtx);
        if(me->title) {
            free(me->title);
            me->title = NULL;
        }
        pthread_mutex_unlock(&me->title_mtx);
        s = get_title(me, me->filename);
        if(s) me->callback(&me->pl, CB_TITCHANGE, me->cb_param, s);
    }
    mad_header_finish(&header);
    mad_stream_finish(&stream);
    if(inbuf.data)
        free(inbuf.data);
    if(fd >= 0)
        close(fd);
    /* wake up other threads waiting for us */
    pthread_mutex_lock(&scan_thread_mtx);
    me->scan_thr_running--;
    pthread_cond_broadcast(&term_scan_cond);
    pthread_mutex_unlock(&scan_thread_mtx);
    DBGMSG("scan_thread for %s ended\n", fn);
    free(fn);
    return NULL;
}

static inline void wait_for_scan_thread(player_mad* me)
{
/*  if(pthread_equal(pthread_self(), me->dec_thr)) */
/*         pthread_mutex_unlock(&scan_thread_mtx); */
    DBGMSG("waiting for scan thread, id %d\n", me->scan_thr);
    if(!pthread_equal(pthread_self(), me->dec_thr))
        pthread_mutex_lock(&scan_thread_mtx);
    while(me->scan_thr_running)
        pthread_cond_wait(&term_scan_cond, &scan_thread_mtx);
    pthread_mutex_unlock(&scan_thread_mtx);
#if 0
    pthread_join(me->scan_thr, NULL);
#endif
    me->scan_thr = (pthread_t) 0;
}

static inline void term_scan_thread(player_mad* me)
{
    me->abort_scan++;
    wait_for_scan_thread(me);
    me->abort_scan--;
}

static inline void term_decoder_thread(player_mad* me)
{
#if 0
    pthread_mutex_lock(&me->dec_msg_mtx);
    me->dec_msg_tx = MSG_TERM;
    pthread_cond_broadcast(&me->dec_msg_cond); /* if it's in pause */
    pthread_mutex_unlock(&me->dec_msg_mtx);
    if(pthread_join(me->dec_thr, NULL) == 0)
        me->dec_thr = (pthread_t) 0;
#else
    send_dec_msg(me, MSG_TERM, dont_wait);
    thrcomm_waitstop(&me->dec_thr_chn);
/*     pthread_join(me->dec_thr, NULL); */
#endif
}

/* decoder thread  */

/* return 1 to continue or 0 to terminate  */
int handle_eof(player_mad* me)
{
    int ret = 0;
    if(me->callback) {
        pthread_mutex_unlock(&scan_thread_mtx);
        ret = me->callback(&me->pl, CB_SONGCHANGE, me->cb_param, NULL);
        pthread_mutex_lock(&scan_thread_mtx);
    }
    if(ret) {
#if 0
        pthread_mutex_lock(&me->dec_msg_mtx);
        if(me->dec_msg_tx == MSG_NONE)
            me->dec_msg_tx = MSG_PAUSE;
        pthread_mutex_unlock(&me->dec_msg_mtx);
#else
/*      auto_send_msg(me, MSG_PAUSE); */
#endif
        return 1;
    } else
        return 0;
#if 0
    if(me->dec_msg_tx != MSG_SONGCHANGE)
        return 0;
    else { /* callback loaded the next song */
        return 1;
    }
#endif
}

typedef enum what_to_do {
    do_nothing = 0, /* ignore error */
    terminate,      /* terminate decoder */
    refill_buf,     /* re-fill input buffer and retry */
    skip_frame,     /* skip to the next frame */
    mute_frame      /* render this frame to silence */
} what_to_do;

what_to_do handle_decoder_error(player_mad* me, struct mad_stream* s,
                                /*buffer* buf, */char* src)
{
    what_to_do ret = skip_frame;
    int show_err = 1;
    int tagsize;
    switch(s->error) {
    case MAD_ERROR_LOSTSYNC:
        /* ignore LOSTSYNC due to ID3 tags */
        tagsize = id3_tag_query (s->this_frame, s->bufend - s->this_frame);
        if (tagsize > 0) {
            DBGMSG("decoder: skipping ID3 tag\n");
            mad_stream_skip (s, tagsize);
        } else
            ret = skip_frame;
        break;

    case MAD_ERROR_BADDATAPTR:
        break;

    case MAD_ERROR_BADCRC:
        ret = mute_frame;
        break;

    default:
        if (!MAD_RECOVERABLE(s->error))
            ret = terminate;
        break;
    }
    if(show_err) {
        pthread_mutex_unlock(&scan_thread_mtx);
        send_error_msg(me, "%s (frame %d): %s", src, me->pos.frames,
                       mad_stream_errorstr(s));
        pthread_mutex_lock(&scan_thread_mtx);
    }

    return ret;
}

void* decoder_thread(void* pme)
{
    player_mad* me = (player_mad*) pme;
    decoder dec;
    buffer inbuf = { NULL, 0, NULL, 0 };
    int out_opened = 0;
    long last_frame_size = buffer_size; /* fill buffer the first time */
    mad_timer_t elapsed = mad_timer_zero;
    int ret;
    int done = 0, song_finished = 0;
    output_fmt outfmt = { 0, 0, 0 }, realfmt;

#if 1
    thrcomm_listen(&me->dec_thr_chn);
#else
    me->dec_thr_running = 1;
#endif
    DBGMSG("decoder_thread started\n");

#ifdef HAVE_MMAP
    if(me->mmap) {
        pthread_mutex_lock(&me->infile_mtx);
        assert(me->buffer);
        mad_stream_buffer(&dec.stream, me->buffer, me->flen);
        pthread_mutex_unlock(&me->infile_mtx);
    } else {
#endif
        /* allocate input buffer */
        inbuf.data = malloc(buffer_size);
        if(!inbuf.data) {
            send_error_msg(me, "decoder: %s", strerror(ENOMEM));
            goto end;
        }
        inbuf.ptr = inbuf.data;
        inbuf.size = buffer_size;
        inbuf.len = 0;
        decoder_init(&dec);
    } 


    pthread_mutex_lock(&scan_thread_mtx);

    while(!done) {
#ifdef HAVE_MMAP
        if(me->mmap) {
            if(dec.stream.error == MAD_ERROR_BUFLEN) {
                me->eof = 1;
                if(!handle_eof(me))
                    break;
                else
                    song_finished = 1;
            }
        } else {
#endif
            if(dec.stream.next_frame) {
                inbuf.ptr = dec.stream.next_frame;
                inbuf.len = inbuf.data + inbuf.size - inbuf.ptr;
            }
            if(me->eof) {
                if(!handle_eof(me))
                    break;
                else
                    song_finished = 1;
            } else {
                pthread_mutex_lock(&me->infile_mtx);
                ret = refill_buffer(&dec.stream, &inbuf, me->infd, &me->eof);
                pthread_mutex_unlock(&me->infile_mtx);
                if(ret == 0) {
                    send_error_msg(me, "decoder (read(%s)): %s", me->filename,
                                   strerror(errno));
                    break;      
                }
            }

#ifdef HAVE_MMAP
        }
#endif
        if(song_finished)
            thrcomm_wait(&me->dec_thr_chn); /* wait messages */
#if 1
        while(!thrcomm_pending(&me->dec_thr_chn)) {
#else
        while(me->dec_msg_tx == MSG_NONE) {
#endif
            int madret;
            ret = do_nothing;
            if(me->mmap) pthread_mutex_lock(&me->infile_mtx);           
            madret = mad_header_decode(&dec.frame.header, &dec.stream);
            if(me->mmap) pthread_mutex_unlock(&me->infile_mtx);         
            if(madret < 0) {
                if(dec.stream.error == MAD_ERROR_BUFLEN)
                    break;
                ret = handle_decoder_error(me, &dec.stream, /*&inbuf,*/
                                           "mad_header_decode");
                if(ret == refill_buf) break;
                if(ret == terminate) { done = 1; break; }
            }

            mad_timer_add(&elapsed, dec.frame.header.duration);
            /* >> TODO: fix elapses with bytes in buffer not played */
            pthread_mutex_lock(&me->pos_mtx);
            me->pos.frames++;
            me->pos.seconds = elapsed.seconds +
                (float)elapsed.fraction / MAD_TIMER_RESOLUTION;
            pthread_mutex_unlock(&me->pos_mtx);
            if(ret == skip_frame)
                continue;


            if(ret != mute_frame) {
                if(me->mmap) {
                    pthread_mutex_lock(&me->infile_mtx);
#if 1
                    if(thrcomm_pending(&me->dec_thr_chn))
#else
                    if(me->dec_msg_tx != MSG_NONE)
#endif
                    {
                        pthread_mutex_unlock(&me->infile_mtx);
                        break; /* intermediate check, me->buffer could be
                                  changed */
                    }
                }
                madret = mad_frame_decode(&dec.frame, &dec.stream);
                if(me->mmap) pthread_mutex_unlock(&me->infile_mtx);         
                if(madret < 0) {
                    if(dec.stream.error == MAD_ERROR_BUFLEN)
                        break;
                    ret = handle_decoder_error(me, &dec.stream, /*&inbuf,*/
                                               "mad_frame_decode");
                    if(ret == refill_buf) break;
                    else if(ret == skip_frame) continue;
                    if(ret == terminate) { done = 1; break; }
                }
            }

            if(me->callback) {
                pthread_mutex_unlock(&scan_thread_mtx);
                me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
                pthread_mutex_lock(&scan_thread_mtx);
            }
            last_frame_size = dec.stream.next_frame - dec.stream.this_frame;

            if(ret != mute_frame) {
                int channels, rate;
                mad_synth_frame(&dec.synth, &dec.frame);
                /* dynamically change output device parameters, so we can handle
                   very strange streams (such as those obtained by erroneous
                   cut'n'paste or like) and seamlessly continue decoding from one
                   song to the next (resetting the device only if needed).
                */
                channels = dec.synth.pcm.channels;
                rate = dec.synth.pcm.samplerate;
                if(!out_opened || outfmt.chan != channels ||
                   (outfmt.rate != rate)) {
                    outfmt.rate = dec.synth.pcm.samplerate;
                    outfmt.chan = channels;
                    outfmt.bits = MAD_F_FRACBITS + 1;
                    if(!output_open(&outfmt, &realfmt)) {
                        send_error_msg(me, "decoder: unable to open output device");
                        done = 1;
                        break;
                    }
                    out_opened = 1;
                }
                {
                    void* chns[2];
                    int bufavail, bufsz, hurry = 0;
                    chns[0] = &dec.synth.pcm.samples[0];
                    chns[1] = &dec.synth.pcm.samples[1];
                    /* let scan_thread run only if this call to
                       output_write_chans will block */
                    if(!me->scan_thr_running || (output_bufinfo(&bufsz, &bufavail)
                     /* && bufavail >= dec.synth.pcm.length))*/
                     && bufavail*4 < bufsz) /* this is a hack - to be removed */)
                        hurry = 1;
                    if(!hurry)
                        pthread_mutex_unlock(&scan_thread_mtx);
                    output_write_chans(chns, dec.synth.pcm.length);
                    if(!hurry)
                      pthread_mutex_lock(&scan_thread_mtx);
                }
            } else { /* play silence */
                /* >>>TODO: maybe we could play last frame instead
                   (as madplay does) under certain conditions */
#if 0 /* hmmm... better just skip, dunno why ... */
                long nsamp = mad_timer_count(dec.frame.header.duration,
                                             dec.frame.header.samplerate);
                DBGMSG("frame muted: playing ~%d msec of silence\n",
                       mad_timer_count(dec.frame.header.duration,
                                       MAD_UNITS_MILLISECONDS));
                if(!setup_outbuf(&outbuf, &outbuf_sz, nsamp * channels)) {
                    send_error_msg(me, "decoder: ", strerror(ENOMEM));
                    done = 1;
                    break;
                }
                memset(outbuf, 0, nsamp * channels * 4);
                pthread_mutex_unlock(&scan_thread_mtx);
                output_write(outbuf, dec.synth.pcm.length);
                pthread_mutex_lock(&scan_thread_mtx);
#endif
            }
            /* heuristically guess (using last frame size) if we need to
               refill input buffer */
            if(inbuf.len < 2*last_frame_size)
                break;
        }
        /* check messages */
        while(thrcomm_pending(&me->dec_thr_chn)) {
            int ack = ACK_DONE;
            decmsg *m = thrcomm_recv(&me->dec_thr_chn);
            switch(m->msg) {
            case MSG_TERM:
                done = 1;
                break;
            case MSG_SEEK: {
                song_pos *seek = &((decmsg_seek*)m)->seek;
                long new_pos;
                output_stop(1, 0);
                /* calculate new address into the file */
                if(me->scan_thr_running) {
                    /* wait for song_pos to be available */
                    wait_for_scan_thread(me);
                }
                if(me->len.frames > 0) { /* else don't seek :-| */
                    pthread_mutex_lock(&me->pos_mtx);
                    new_pos = (int)(me->flen *
                                    (double)seek->frames / me->len.frames);
                    pthread_mutex_unlock(&me->pos_mtx);
                    /* >>>TODO: implement a more accurate seeking */
#ifdef HAVE_MMAP
                    if(me->mmap) {
                        assert(me->buffer);
                        mad_stream_buffer(&dec.stream, me->buffer + new_pos,
                                          me->flen - new_pos);
                        ret = 1;
                    } else {
#endif
                        pthread_mutex_lock(&me->infile_mtx);
                        lseek(me->infd, new_pos, SEEK_SET);
                        inbuf.ptr = inbuf.data;
                        inbuf.len = 0;
                        ret = refill_buffer(&dec.stream, &inbuf, me->infd, &me->eof);
                        pthread_mutex_unlock(&me->infile_mtx);                      
                        if(!ret) {
                            send_error_msg(me, "decoder (read(%s)): %s",
                                           me->filename, strerror(errno));
                            done = 1; /* cleanly terminate thread */
                        }
#ifdef HAVE_MMAP
                    }
#endif
                    if(ret != 0) {
                        me->pos.frames = seek->frames;
                        /* >>>TODO: this is very inaccurate (especially for VBR) */
                        if(seek->seconds > 0)
                            me->pos.seconds = seek->seconds;
                        else
                            me->pos.seconds = me->len.seconds * 
                                (double)seek->frames / me->len.frames;
                        mad_timer_set(&elapsed, 0, (unsigned long) me->len.seconds *
                                      seek->frames, me->len.frames);
                        if(!me->mmap)
                            /* force a buffer refill */
                            dec.stream.error = MAD_ERROR_BUFLEN;
                        dec.stream.sync = 0;
                        /* skip two frames */
                        if(mad_frame_decode(&dec.frame, &dec.stream) > 0)
                            mad_frame_decode(&dec.frame, &dec.stream);
                    } else
                        ack = ACK_ERROR;                    
                } /* end if(song_len is available) */
                else
                    ack = ACK_ERROR;
                break;
            }
            case MSG_PAUSE:
                output_pause();
                if(m->ack) {
                    *m->ack = ACK_PAUSE;
                    thrcomm_ack(&me->dec_thr_chn);
                }
                pthread_mutex_unlock(&scan_thread_mtx);
                /* wait for some messages */
                thrcomm_wait(&me->dec_thr_chn);
                pthread_mutex_lock(&scan_thread_mtx);
                free(m);
                continue; /* process the new message */

            case MSG_SONGCHANGE:
                /* it the last song wasn't finished, flush output buffer, else
                   let it there so there's no gap between songs */
                if(!song_finished)
                    output_stop(1, 0);
                else
                    song_finished = 0;
                /* reset_decoder  */
                decoder_finish(&dec);
                decoder_init(&dec);
                mad_timer_reset(&elapsed);
                if(me->callback) {
                    pthread_mutex_unlock(&scan_thread_mtx);
                    me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
                    pthread_mutex_lock(&scan_thread_mtx);
                }
#ifdef HAVE_MMAP
                if(me->mmap) {
                    assert(me->buffer);
                    mad_stream_buffer(&dec.stream, me->buffer, me->flen);
                } else {
#endif      
                    inbuf.ptr = inbuf.data;
                    inbuf.len = 0;
                    pthread_mutex_lock(&me->infile_mtx);
                    ret = refill_buffer(&dec.stream, &inbuf, me->infd, &me->eof);
                    pthread_mutex_unlock(&me->infile_mtx);
                    if(!ret) {
                        send_error_msg(me, "decoder (reading file): %s",
                                       strerror(errno));
                        done = 1; /* cleanly terminate thread */
                        ack = ACK_ERROR;
                    }
#ifdef HAVE_MMAP
                }
#endif
                break;
            default:
            case MSG_RESUME:
            case MSG_STOP: /* not used -- ignored */
                break;
            }
#if 1
            if(m->msg != MSG_PAUSE && m->ack) {
                *m->ack = ack;
                thrcomm_ack(&me->dec_thr_chn);
            }
            free(m);
        }
#else
            /* acknowledge that message was handled */
            pthread_cond_broadcast(&me->dec_msg_cond);
        }
        pthread_mutex_unlock(&me->dec_msg_mtx);
#endif

    } /* end of main loop */
 end:

    pthread_mutex_unlock(&scan_thread_mtx);

    if(out_opened) {
        output_stop(1, 0);
        output_close();
    }
    decoder_finish(&dec);
    if(inbuf.data)
        free(inbuf.data);
    me->pl.status = st_stop;
    pthread_mutex_lock(&me->pos_mtx);
    me->pos.frames = 0;
    me->pos.seconds = 0.0;
    pthread_mutex_unlock(&me->pos_mtx);
    if(me->callback) {
        me->callback(&me->pl, CB_UPDATE, me->cb_param, NULL);
        me->callback(&me->pl, CB_STATUS, me->cb_param, NULL);
    }
#if 1
    thrcomm_stop(&me->dec_thr_chn);
#else
    me->dec_thr_running = 0;
#endif
    DBGMSG("decoder_thread ended\n");
    return NULL;
}

/* read id3 tag */

static const char* const separator = "  "; /* replaces '\n' */

void insert_tag_comment(struct id3_tag const* tag, string* dest)
{
    unsigned int i;
    struct id3_frame* frame;
    id3_ucs4_t const *ws;
    id3_latin1_t *s, *p;
    /* it's more efficient to scan ourselves instead of using
       id3_tag_findframe repeatedly */
    for(i = 0; i < tag->nframes; i++) {
        frame = tag->frames[i];
        if(strcmp(frame->id, ID3_FRAME_COMMENT) == 0) {
            ws = id3_field_getstring(&frame->fields[2]);
            if (*ws == '\0') {
                ws = id3_field_getfullstring(&frame->fields[3]);
                s = id3_ucs4_latin1duplicate(ws);
                assert(s);
                p = s;
                while(1) {
                    id3_latin1_t* t = strchr(p, '\n');
                    if(!t) break;
                    if(p != s)
                        string_append(dest, separator);
                    string_nappend(dest, p, t - p);
                    p = t + 1;
                }
                string_append(dest, p);
                free(s);
            }
        }
    }
}

void insert_tag_genre(struct id3_frame const* frame, string* dest)
{
    union id3_field* field;
    id3_latin1_t* s;
    id3_ucs4_t const* ws;
    field = &frame->fields[1];
    ws = id3_field_getstrings(field, 0);
    if(ws) {
        if(*ws >= '0' && *ws <= '9')
            ws = id3_genre_name(ws);
        s = id3_ucs4_latin1duplicate(ws);
        if(s) {
            string_append(dest, s);
            free(s);
        }
    }
}

void insert_tag_strlist(union id3_field const* field, string* dest)
{
    id3_ucs4_t const* ws;
    id3_latin1_t* s;
    int ns = id3_field_getnstrings(field), i;
    for(i = 0; i < ns; i++) {
        ws = id3_field_getstrings(field, i);
        /* no internationalization support yet */
        s = id3_ucs4_latin1duplicate(ws);
        if(i > 0)
            string_append(dest, separator);
        string_append(dest, s);
        free(s);
    }
}

void insert_tag_strlist_latin1(union id3_field const* field, string* dest)
{
    id3_latin1_t *s;
    int ns = field->latin1list.nstrings, i;
    for(i = 0; i < ns; i++) {
        if(i > 0)
            string_append(dest, separator);
        s = field->latin1list.strings[i];
        if(s) string_append(dest, s);
    }
}

string* insert_tag_value(struct id3_tag const* tag, const char* id, string* dest)
{
    struct id3_frame const *frame;
    union id3_field const *field;

    if(strcmp(id, ID3_FRAME_COMMENT) == 0) {
        insert_tag_comment(tag, dest);
    } else {
        frame = id3_tag_findframe(tag, id, 0);
        if (frame == 0)
            return dest; /* don't insert nothing */
        if(strcmp(id, ID3_FRAME_GENRE) == 0) {
            insert_tag_genre(frame, dest);
        } else if(frame->nfields == 2 && /* check if this is a text frame */
                  frame->fields[1].type == ID3_FIELD_TYPE_STRINGLIST) {
            field = &frame->fields[1]; /* field 1 is the string list for text frames */
            insert_tag_strlist(field, dest);
        } else {
            int inserted = 0;
            unsigned int i;
            id3_latin1_t* s;
            /* do something */
            DBGMSG("Should display tag %s, which isn't of type text\n", id);
            for(i = 0; i < frame->nfields; i++) {
                field = &frame->fields[i];
                if(inserted && i != 0)
                    string_append(dest, separator);
                inserted = 1;
                switch(field->type) {
                case ID3_FIELD_TYPE_LATIN1:
                case ID3_FIELD_TYPE_LATIN1FULL:
                    if(field->latin1.ptr)
                        string_append(dest, field->latin1.ptr);
                    else inserted = 0;
                    break;
                case ID3_FIELD_TYPE_LATIN1LIST:
                    insert_tag_strlist_latin1(field, dest);
                    break;
                case ID3_FIELD_TYPE_STRING:
                case ID3_FIELD_TYPE_STRINGFULL:
                    if(field->type == ID3_FIELD_TYPE_STRING)
                        s = id3_ucs4_latin1duplicate(id3_field_getstring(field));
                    else
                        s = id3_ucs4_latin1duplicate(id3_field_getfullstring(field));
                    if(s) string_append(dest, s);
                    else inserted = 0;
                    break;
                case ID3_FIELD_TYPE_STRINGLIST:
                    insert_tag_strlist(field, dest);
                    break;

                case ID3_FIELD_TYPE_LANGUAGE:
                case ID3_FIELD_TYPE_FRAMEID:
                case ID3_FIELD_TYPE_DATE:
                    string_append(dest, field->immediate.value);
                    break;

                case ID3_FIELD_TYPE_INT8:
                case ID3_FIELD_TYPE_INT16:
                case ID3_FIELD_TYPE_INT24:
                case ID3_FIELD_TYPE_INT32: {
                    char value[11];
                    snprintf(value, sizeof(value), "%ld", field->number.value);
                    string_append(dest, value);
                    break;
                }
                default:
                    inserted = 0;
                    break;
                }
            }
        }
    }
    return dest;
}

/* mad_title: a string with the following symbols to be substituited:
   (must be lowercase)
   %f => filename (without path and extension)
   %d => filename path (without the trailing /)
   %e => filename extension (should be 'mp3', right?)
   %p => complete filename
   %t => song title
   %a => author name
   %l => album name
   %g => genre
   %y => year
   %n => track number
   %c => comments
   (must be uppercase)
   %XXXX% => the 'XXXX' id3v2 tag
   %% => the '%' character
*/
char* make_title(player_mad* me, const char* format,
                 struct id3_tag const* tag, const char* fn)
{
    string s;
    const char* p = format, *t, *q;
    char *tit;
    int i;
    if(!p) return NULL;
    string_init(&s);
    while(1) {
        t = strchr(p, '%');
        if(t) {
            string_nappend(&s, p, t - p);
            switch(*(++t)) {
            case '%':
                string_append(&s, "%");
                break;
            case 'f': /* name without path nor ext */
                q = getname(fn);
                string_nappend(&s, q, getext(q) - q);
                break;
            case 'd':
                q = getname(fn);
                if(q != fn)
                    string_nappend(&s, fn, q - fn - 1);
                break;
            case 'p':
                string_append(&s, fn);
                break;
            case 'e':
                q = getext(fn);
                if(*q != '\0')
                    string_append(&s, getext(fn) + 1);
                break;
            case 'b': {
                char value[11];
                if(me->bitrate < 0) {
                    string_append(&s, "(unknown)");
                } else {
                    snprintf(value, sizeof(value), "%lu",
                             (me->bitrate + 500) / 1000);
                    string_append(&s, value);
                }
                break;
            }               
            case 't':
                insert_tag_value(tag, ID3_FRAME_TITLE, &s); break;
            case 'a':
                insert_tag_value(tag, ID3_FRAME_ARTIST, &s); break;
            case 'l':
                insert_tag_value(tag, ID3_FRAME_ALBUM, &s); break;
            case 'g':
                insert_tag_value(tag, ID3_FRAME_GENRE, &s); break;
            case 'y':
                insert_tag_value(tag, ID3_FRAME_YEAR, &s); break;
            case 'n':
                insert_tag_value(tag, ID3_FRAME_TRACK, &s); break;
            case 'c':
                insert_tag_value(tag, ID3_FRAME_COMMENT, &s); break;
            default:
                /* isupper('\0') == 0 so this won't cause segfaults */
                for(i = 0; i < 4; i++)
                    if(!isupper(t[i]) && !isdigit(t[i]))
                        break;
                if(i == 4 && t[4] == '%') {
                    char id[5] = { t[0], t[1], t[2], t[3], '\0' };
                    insert_tag_value(tag, id, &s);
                    t += 4;
                } else {
                    send_error_msg(me, "invalid format specifier %%%c\n", *t);
                }
                break;
            }
            p = t + 1;
        } else { /* no % found */
            string_append(&s, p);
            break;
        }
    } /* end while */
    tit = string_getbytes(&s);
    string_finish(&s);
    return tit;
}

/* mad player methods */

int mad_init(player_mad* me)
{
    me->pl = player_mad_tmpl;
    me->callback = NULL;
    me->pos.frames   = 0;
    me->pos.seconds  = 0.0;
    me->len.frames   = 0;
    me->len.seconds  = 0.0;
    me->bitrate = -1;
    pthread_mutex_init(&me->pos_mtx, NULL);
    me->filename = NULL;
    me->title = NULL;
    pthread_mutex_init(&me->title_mtx, NULL);
    me->infd = 0;
    me->flen = 0;
    me->eof = 0;
    me->mmap = 0;
#ifdef HAVE_MMAP
    me->buffer = NULL;
#endif
    pthread_mutex_init(&me->infile_mtx, NULL);
    me->scan_thr = me->dec_thr = (pthread_t) 0;
    me->scan_thr_running /*= me->dec_thr_running*/ = 0;
    me->abort_scan = 0;
#if 0
    pthread_mutex_init(&me->dec_msg_mtx, NULL);
    pthread_cond_init(&me->dec_msg_cond, NULL);
    me->dec_msg_tx = MSG_NONE;
    me->dec_msg_rx = ACK_DONE;
#else
    thrcomm_init(&me->dec_thr_chn);
#endif
    pthread_mutex_init(&me->my_mtx, NULL);
    return 0;
}

void mad_finish(struct player* pme) /* destructor */
{
    player_mad* me = (player_mad*) pme;
    if(me->scan_thr > 0)
        term_scan_thread(me);
    if(me->dec_thr > 0)
        term_decoder_thread(me);
        
#ifdef HAVE_MMAP
    if(me->mmap && me->buffer) {
        munmap(me->buffer, me->flen);
        me->buffer = NULL;
        me->mmap = 0;
    }
#endif
    if(me->infd > 0) {
        close(me->infd);
        me->infd = 0;
    }
    thrcomm_finish(&me->dec_thr_chn);
    /* these two aren't strictly needed, useful for debugging */
    me->flen = 0;
    me->eof = 0;
    /* free allocated memory */
    if(me->title) free(me->title);
    if(me->filename) free(me->filename);
}

static inline int this_is_dec_thr(player_mad* me)
{
    return pthread_equal(pthread_self(), me->dec_thr);
}

static inline int start_decoder_thread(player_mad* me)
{
    int ret = pthread_create(&me->dec_thr, NULL, decoder_thread, me);
    if(ret == 0)
        pthread_detach(me->dec_thr); /* don't use pthread_join anymore */
    return ret;
}
 
static inline int start_scan_thread(player_mad* me)
{
    int ret;
    pthread_mutex_lock(&scan_thread_mtx);
    if(me->scan_thr_running) {
        pthread_mutex_unlock(&scan_thread_mtx);
        term_scan_thread(me);
        pthread_mutex_lock(&scan_thread_mtx);
    }
/*     me->abort_scan = 0; */
    ret = pthread_create(&me->scan_thr, NULL, scan_thread, me);
    if(ret == 0) {
        me->scan_thr_running++;
        pthread_detach(me->scan_thr); /* don't use pthread_join anymore */
        DBGMSG("scan_thread started, id %d\n", me->scan_thr);
    }
    pthread_mutex_unlock(&scan_thread_mtx);
    return ret;
}
 
/* player methods */

/* test if the file is in the correct format for this player */
int mad_test(struct player* pme, const char* fn)
{
    /*     player_mad* me = (player_mad*) pme; */
    int is_mp3 = 0;
    /* do some checks to see if this is an mp3 file, if they all fail
       try to load first header, then give up */
    /* this code is taken from xmms-mad (http://xmms-mad.sourceforge.net/)
       (a little bit reformatted and readapted) */
    int32_t check = 0L;
    int fin = open (fn, O_RDONLY);
    (void) pme;
    if (fin >= 0 && read (fin, &check, 4) == 4) {
        /* If first two bytes are a sync header or three bytes are "ID3" */
        if ( ((check & 0xf0ff) == 0xf0ff) || ((check & 0xffffff) == 0x334449)) {
            is_mp3 = 1;
        } else if (check == 0x46464952) { /* RIFF RMP3 stream */    
            lseek (fin, 4, SEEK_CUR);
            if (read (fin, &check, 4) == 4 && check == 0x33504d52)
                is_mp3 = 1;
        }
    }
    /* end xmmd-mad code */
    if(!is_mp3 && fin >= 0) {
        /* last resort: try to load first frame, skipping any junk
           eventually found in the first 4 kilobytes */
        enum { TEST_BUF_SIZE = 4096 };
        char* buffer = malloc(TEST_BUF_SIZE);
        struct mad_stream stream;
        struct mad_header header;
        lseek(fin, 0L, SEEK_SET);
        if(buffer && read(fin, buffer, TEST_BUF_SIZE) == TEST_BUF_SIZE) {
            mad_stream_init(&stream);
            mad_header_init(&header);
            mad_stream_buffer(&stream, buffer, TEST_BUF_SIZE);
            if((mad_header_decode(&header, &stream) != -1)
               || (stream.error == MAD_ERROR_BUFLEN))
                is_mp3 = 1;
            else if(stream.error == MAD_ERROR_LOSTSYNC) {
                /* skip junks at the beginning of the file */
                if(mad_header_decode(&header, &stream) >= 0)
                    is_mp3 = 1;
            }
            mad_header_finish(&header);
            mad_stream_finish(&stream);
        }
        if(buffer) free(buffer);
    }
    if(fin >= 0)
        close(fin);
    return is_mp3;
}

char* get_title(player_mad* me, const char* fn)
{
    struct id3_file* file = NULL;
    struct id3_tag* tag = NULL;
    char* s = NULL;
    pthread_mutex_lock(&me->title_mtx);
    if(!me->title) {
        file = id3_file_open(fn, ID3_FILE_MODE_READONLY);
        if(file) {
            tag = id3_file_tag(file);
            if(tag && tag->nframes > 0)
                s = make_title(me, cfg_title, tag, fn);
            else    
                s = make_title(me, cfg_title_notag, tag, fn);
            me->title = s;
            id3_file_close(file);
        } else {
            fprintf(stderr, "get_title: error opening %s for reading.\n\t%s\n",
                    fn, strerror(errno));
        }
    } else
        s = me->title;
    pthread_mutex_unlock(&me->title_mtx);
    return s;
}

/* setup `filename' for playing and, if current status is st_play,
   or 'play' is != 0, start playing it.
   Stores in 'title' (if != NULL) a malloc'ed buffer with the title to show
   (may be something like 'title - album - artist', not just the song title).
   It's your responsibility to free the title buffer (if requested).
   Use me->setup(me, NULL, NULL); to release everything.
*/
int mad_setup(struct player* pme, const char* fn, char ** title, int play)
{
    player_mad* me = (player_mad*) pme;
    int need_to_synchronize = (thrcomm_running(&me->dec_thr_chn)
                               && !this_is_dec_thr(me));
    struct stat st;
    int samefile = 0;
    int ret = 1;
    /* Open file and load ID3.
       If dec_thr is valid, set MSG_PAUSE before setup, and release it after,
       else if play == 1 call mad_play.
    */
    if(stat(fn, &st) < 0)
        return 0; /* file does not exist */

    pthread_mutex_lock(&me->my_mtx); /* synchronize calls to mad_setup */
    DBGMSG("mad_setup(%s): started\n", fn);

    pthread_mutex_lock(&me->infile_mtx);
    if(me->infd >= 0 && me->filename &&
       strcmp(me->filename, fn) == 0 &&
       st.st_mtime == me->mtime)
        samefile = 1; /* re-setting up same file */
    pthread_mutex_unlock(&me->infile_mtx);

    if(me->scan_thr_running && !samefile) {
        DBGMSG("mad_setup(%s): terminating scan_thread\n", fn);
        term_scan_thread(me);
    }

    if(need_to_synchronize)
        mad_pause(&me->pl);
    if(!samefile) {
        pthread_mutex_lock(&me->infile_mtx);
        if(me->infd) {
#ifdef HAVE_MMAP
            if(me->mmap) {
                munmap(me->buffer, me->flen);
                me->mmap = 0;
                me->buffer = NULL;
            }
#endif
            close(me->infd);
            me->infd = -1;
        }
        me->flen = st.st_size;
        me->mtime = st.st_mtime;
        me->infd = open(fn, O_RDONLY);
        if(me->infd < 0) {
            send_error_msg(me, "mad_setup (open(%s)): %s", fn, strerror(errno));
            pthread_mutex_unlock(&me->my_mtx);
            return 0;
        }
#ifdef HAVE_MMAP
        /* try to mmap the file */
        if(cfg_mmap) {
            me->buffer = mmap(NULL, me->flen, PROT_READ, MAP_PRIVATE, me->infd, 0);
            if((int)me->buffer == -1) {
                me->buffer = NULL;
                me->mmap = 0;
            } else {
                DBGMSG("mad_setup: successfully mmap'ed %s\n", fn);
                me->mmap = 1;
                madvise(me->buffer, me->flen, MADV_SEQUENTIAL);
                /* preload some data  */
                madvise(me->buffer, buffer_size, MADV_WILLNEED);
            }
        } else {
            me->buffer = NULL;
            me->mmap = 0;
        }
#endif              
        pthread_mutex_unlock(&me->infile_mtx);
        
        if(me->filename) free(me->filename);
        me->filename = strdup(fn);
/*      if(!me->filename) return 0; */
        if(me->title) free(me->title);
        me->title = NULL;
        me->eof = 0;

        me->len.frames = -1;
        me->bitrate = -1;
#ifndef PROFILING
        /* scan file in the background */
        DBGMSG("mad_setup(%s): starting scan_thread\n", fn);
        start_scan_thread(me);
#endif
    } else {/* if(!samefile) */
        pthread_mutex_lock(&me->infile_mtx);        
        lseek(me->infd, 0L, SEEK_SET); /* rewind */
        pthread_mutex_unlock(&me->infile_mtx);
    }
    me->pos.frames = 0;
    me->pos.seconds = 0.0;

    if(title) {
        char *s = get_title(me, fn);
        if(s) *title = strdup(s);
        else  *title = NULL;
    }

    if(need_to_synchronize) {
#if 0       
        pthread_mutex_lock(&me->dec_msg_mtx);
        me->pl.status = st_play; /* this was changed calling mad_pause */
        me->dec_msg_tx = MSG_SONGCHANGE; /* inform decoder of the change */
        pthread_cond_broadcast(&me->dec_msg_cond);
        pthread_mutex_unlock(&me->dec_msg_mtx);
#else
        me->pl.status = st_play; /* this was changed calling mad_pause */
        send_dec_msg(me, MSG_SONGCHANGE, dont_wait);
#endif
    } else if(this_is_dec_thr(me)) {
        ret = auto_send_msg(me, MSG_SONGCHANGE);
    } else {
        if(play)
            ret = mad_play(pme);
        else
            me->pl.status = st_stop;
    }
    DBGMSG("mad_setup(%s): finish\n", fn);
    pthread_mutex_unlock(&me->my_mtx); /* synchronize calls to mad_setup */
    return ret;
}

int mad_play(struct player* pme)    /* guess ;) */
{
    player_mad* me = (player_mad*) pme;
    int err;
    if(me->pl.status == st_pause) {
        return mad_pause(pme);
    } else {
#ifdef PROFILING
        /* scan file in the background */
        start_scan_thread(me);
#endif
        if(/*!me->dec_thr_running*/!thrcomm_running(&me->dec_thr_chn)) {
            if((err = start_decoder_thread(me)) != 0) {
                send_error_msg(me, "mad_play (pthread_create): %s",
                               strerror(err));
                return 0;
            }
            /* wait for decoder_thread to start */
            thrcomm_waitlisten(&me->dec_thr_chn);
        }
        me->pl.status = st_play;
    }
    return 1;
}

/* N.B. synchronous playing is implemented to enable PROFILING (because
   it has problems with threads) and nothing else */
int mad_play_sync(struct player* pme)
{
    player_mad* me = (player_mad*) pme;
    if(me->pl.status == st_pause || me->pl.status == st_play) {
        if(!mad_stop(pme))
            return 0;
    }
    if(me->scan_thr > (pthread_t)0)
        term_scan_thread(me);
/*     me->dec_msg_tx = MSG_NONE; */
    scan_thread(me); /* this is to enable profiling of scan thread too */
    me->pl.status = st_play;
    me->dec_thr = pthread_self();

    decoder_thread(me); /* start decoder synchronously */

    me->dec_thr = (pthread_t)0;
    me->pl.status = st_stop;
    return 1;
}

/* 'pause' will pause if playing and resume if already paused
   (and do nothing if stopped) */
int mad_pause(struct player* pme)
{
    player_mad* me = (player_mad*) pme;
    int ret = 1;
    if(thrcomm_running(&me->dec_thr_chn)/*me->dec_thr_running*/) {
        /* pause output immediately and set MSG_PAUSE, or set MSG_RESUME and
           signal dec_cond if already paused */
        if(me->pl.status == st_play) {
#if 1
            if(send_dec_msg(me, MSG_PAUSE, wait_ack) != ACK_PAUSE)
              ret = 0;
#else
            pthread_mutex_lock(&me->dec_msg_mtx);
            me->dec_msg_tx = MSG_PAUSE;
            if(!this_is_dec_thr(me)) {
                pthread_cond_broadcast(&me->dec_msg_cond);
                while(me->dec_msg_rx != ACK_PAUSE && me->dec_msg_rx != ACK_ERROR)
                    pthread_cond_wait(&me->dec_msg_cond, &me->dec_msg_mtx);
                if(me->dec_msg_rx == ACK_ERROR)
                    ret = 0;
            }
            pthread_mutex_unlock(&me->dec_msg_mtx);
#endif
            me->pl.status = st_pause;
        } else if(me->pl.status == st_pause) {
            /* hmmm... this cannot be decoder thread, right? */
#if 1
            send_dec_msg(me, MSG_RESUME, dont_wait);
#else
            pthread_mutex_lock(&me->dec_msg_mtx);
            me->dec_msg_tx = MSG_RESUME;
            pthread_cond_broadcast(&me->dec_msg_cond);
            pthread_mutex_unlock(&me->dec_msg_mtx);
#endif
            me->pl.status = st_play;
        }
    } else {
        me->pl.status = st_stop;
    }
    return ret;
}

int mad_stop(struct player* pme)    /* guess ;) */
{
    int ret = 1;
    player_mad* me = (player_mad*) pme;
    if(me->pl.status != st_stop) {
        if(thrcomm_running(&me->dec_thr_chn)/*me->dec_thr_running*/) {
            if(!this_is_dec_thr(me)) {
#if 1
                term_decoder_thread(me);
#else
                int err;
#if 1
                send_dec_msg(me, MSG_TERM, dont_wait);
#else
                pthread_mutex_lock(&me->dec_msg_mtx);
                me->dec_msg_tx = MSG_TERM;
                pthread_cond_broadcast(&me->dec_msg_cond);
                pthread_mutex_unlock(&me->dec_msg_mtx);
#endif
                err = pthread_join(me->dec_thr, NULL);
                if(err != 0) {
                    send_error_msg(me, "mad_stop (pthread_join): %s",
                                   strerror(err));
                    ret = 0;
                }
#endif
            } else {
#if 1
                auto_send_msg(me, MSG_TERM);
#else
                me->dec_msg_tx = MSG_TERM;
#endif
            }
        }
        me->pl.status = st_stop;
    }
    return ret;
}


static int do_jump(player_mad* me, long frames, float seconds)
{
    song_pos seek;
    if(me->pl.status != st_play)
        if(!mad_play(&me->pl))
            return 0;
#if 1
    seek.frames = frames;
    seek.seconds = seconds;
    if(!this_is_dec_thr(me)) {
        if(send_dec_seek(me, &seek, wait_ack) != ACK_DONE)
            return 0;
    } else
        auto_send_seek(me, &seek);
#else
    /* set dec_seek and MSG_SEEK, eventually start playing */
    pthread_mutex_lock(&me->dec_msg_mtx);
    me->dec_seek.frames = frames;
    me->dec_seek.seconds = seconds;
    me->dec_msg_tx = MSG_SEEK;
    if(!this_is_dec_thr(me)) {
        me->dec_msg_rx = ACK_NONE;
        while(me->dec_msg_rx != ACK_DONE && me->dec_msg_rx != ACK_ERROR) {
            pthread_cond_wait(&me->dec_msg_cond, &me->dec_msg_mtx);
        }
    }
    pthread_mutex_unlock(&me->dec_msg_mtx);
#endif
    return 1;
}

/* jumps 'f' frames from the beginning */
int mad_jump_frames(struct player* pme, long f)
{
    player_mad* me = (player_mad*) pme;
    float seconds;
    if(me->scan_thr_running)
        wait_for_scan_thread(me);
    if(me->len.seconds <= 0) seconds = -1.0;
    else seconds = (float)(f * me->len.seconds / me->len.frames);
    return do_jump(me, f, seconds);
}

/* jumps 'sec' seconds from the beginning */
int mad_jump_time(struct player* pme, long sec)
{
    player_mad* me = (player_mad*) pme;
    /* >>>TODO: implement a more accurate seeking! */
    int frames;
    if(me->scan_thr_running)
        wait_for_scan_thread(me);
    if(me->len.seconds <= 0) return 0;
    frames = (long)((double)sec * me->len.frames / me->len.seconds);
    return do_jump(me, frames, (float)sec);
}

/* jumps to `position/total' of the total file (useful for progress bar) */
int mad_jump(struct player* pme, long position, long total)
{
    player_mad* me = (player_mad*) pme;
    /* >>>TODO: implement a more accurate seeking! */
    int frames;
    float seconds;
    if(me->scan_thr_running)
        wait_for_scan_thread(me);
    if(me->len.frames <= 0 || me->len.seconds <= 0)
        return 0;
    frames = (long)((double)position * me->len.frames / total);
    seconds = (float)((double)position * me->len.seconds / total);
    return do_jump(me, frames, seconds);
}

/* N.B. jump* methods will start playing if status != st_play */

/* query total length (total number of frames/seconds) */
int mad_get_length(struct player* pme, song_pos* len)
{
    player_mad* me = (player_mad*) pme;
    if(me->scan_thr_running)
        wait_for_scan_thread(me);
    if(me->len.frames > 0) {
        *len = me->len;
        return 1;
    } else {
        len->frames = -1;
        len->seconds = 0;
        return 0;
    }
}

/* query current position */
int mad_get_pos(struct player* pme, song_pos* pos)
{
    player_mad* me = (player_mad*) pme;
    int ret = 1;
    pthread_mutex_lock(&me->pos_mtx);
    if(me->pos.frames >= 0) {
        *pos = me->pos;
    } else {
        pos->frames = 0;
        pos->seconds = 0;
        ret = 0;
    }
    pthread_mutex_unlock(&me->pos_mtx);
    return ret;
}


/* set callback */
void mad_set_callback(struct player* pme, player_cb_func cb, void* param)
{
    player_mad* me = (player_mad*) pme;
    me->callback = cb;
    me->cb_param = param;
}

