/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * Based on wmmp3: Copyright (c)1999 Patrick Crosby <xb@dotfiles.com>.
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * testcfg.c
 *
 * test program to test the configuration file parser (parsecfg.c)
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#include "parsecfg.h"

const char * specs =
"mpg123 = string:mpg123;\n"    /* path to mpg123 executable (searched in $PATH)*/
"mp3ext = +string:mp3;\n"      /* mp3 valid extension(s) (default: .mp3) */
"playlistext = +string:m3u;\n" /* playlist valid extension(s) (default: .m3u) */
"alwaysscroll = boolean:no;\n" /* always scroll title (never stops) */
"sort = string:sorted;\n"      /* how to sort dir files: unsorted, sorted, reversed */
"scrolldelay = real:0.5;\n"    /* scrolling delay in seconds */
/* list of { type, name, spec } where:
 * type is one of: dir, file, list (directory, list file or inline list resp.)
 * name is the name of the list
 * spec is the path of the directory of playlist file (type=dir or type=file),
 * or a list of mp3 files in the case of an inline list.
 */
"lists = +{ string, string, +string };\n"
/* this 'skin' variable is just a test, by now (to test optional values
 * and redefinition of a non-array complex variable).
 * Maybe if il'll add skins i'll do something less braindead.
 * skin = { name, pixmap, [ +{ type, x, y, w, h, [ name, +command ] } ] };
 */
"skin = { string, string, [ +{ string:'button', integer:0, integer:0, \n"
"integer:10, integer:10, [ string:'', +string:'' ] } ] };";

#if 0 /* this function is now in parsecfg.c */
char* get_cfg_err(int err)
{
    switch(err) {
    case CFGERR_NONE:
        return "no error";
    case CFGERR_NOTFOUND:
        return "config file not found in ~/, /usr/local/etc or /etc";
    case CFGERR_SYSTEM:
        return /*strerror(errno)*/"there were system errors";
    case CFGERR_SPECS:
        return "error(s) parsing internal specs";
    case CFGERR_FILE:
        return "error(s) parsing configuration file";
    }
    return "(unknown/invalid error code)";
}
#endif

void show_values_rec(opt_tree_node* n, int depth)
{
    printf("%*c", depth*2, ' '); /* indentation */
#ifndef NDEBUG
    printf("%p ", n);
#endif
    switch(n->type) {
    case nt_null:
        printf("<header>\n");
        break;
    case nt_real:
        printf("real: %f (def = %f)\n", n->val.r, n->def.r);
        break;
    case nt_integer:
        printf("integer: %ld (def = %ld)\n", n->val.i, n->def.i);
        break;
    case nt_boolean:
        printf("boolean: %d (def = %d)\n", n->val.b, n->def.b);
        break;
    case nt_string:
        printf("string: \"%s\" (def = \"%s\")\n", n->val.s, n->def.s);
        break;
    case nt_struct:
        printf("struct:\n");
        break;
    case nt_array:
        printf("array:\n");
        break;
    case nt_optional:
        printf("optional:\n");
        break;
    default:
        printf("something strange: %d\n", n->type);
        break;
    }
    if(n->type > nt_composite) {
        opt_tree_node* nd = n->val.child;
#ifndef NDEBUG      
        if(!nd) {
            fprintf(stderr, "%*cThis node has no children (not even the header!)\n",
                    depth*2, ' ');
            return;
        }
        if(nd->type != nt_null) {
            fprintf(stderr, "%*cThis node has no header!\n",
                    depth*2, ' ');
        }
#endif
        for(; nd != NULL; nd = nd->next) {
#ifndef NDEBUG      
            if(nd->parent != n) {
                printf("%*cThe following node is broken (parent=%p)\n",
                       (depth+1)*2, ' ', nd->parent);
            }
#endif
            show_values_rec(nd, depth+1);
        }
    }
}

void show_values(opt_hash* ht)
{
    opt_hash_ent* ent;
    opt_hash_iterator it;
    ent = opt_hash_start(ht, &it);
    for(; ent != NULL; ent = opt_hash_next(&it)) {
        opt_tree_node* n;
        if(ent->tree->visited)
            printf("%s =\n", ent->varname);
        else
            printf("%s = (default values)\n", ent->varname);

        n = ent->tree->root;
#ifndef NDEBUG      
        if(!n)
            fprintf(stderr, "  The tree has no children (not even the header!)\n");
        if(n->type != nt_null)
            fprintf(stderr, "  This node has no header!\n");
#endif
        
        for(; n != NULL; n = n->next)
            show_values_rec(n, 1);
    }
}

int main(int argc, char** argv)
{
    char* name = "testcfgrc";
    opt_hash ht;
    int err;
    if(argc > 1)
        name = argv[1];
    printf("config name: %s\n", name);
    opt_hash_init(&ht);
    err = load_config_file(name, specs, &ht, CFL_SEARCHWD);

    if(err != CFGERR_NONE)
        fprintf(stderr, "Error loading config file: %s\n", get_cfg_err(err));
    else
        printf("Config file loaded successfully.\n");
    printf("Values loaded are:\n");
    show_values(&ht);
    opt_hash_free(&ht);
    return 0;
}
